#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

""" Module to retrieve the data to plot (both X axis and Y axis). The data source can be a BSD file or a WAGASCI file
or a slow device data. The module is implemented using the Strategy design pattern. Each way of collecting the data
is generically called harvester and corresponds to a different strategy. """

import abc
import collections
import operator
import os
import re
from typing import Optional, List, Any, Union, Tuple, Dict

import numpy
from six import string_types

import wagascianpy.analysis.analysis
import wagascianpy.analysis.beam_summary_data
import wagascianpy.analysis.spill as sp
import wagascianpy.database.db_record
import wagascianpy.database.wagascidb as wgdb
import wagascianpy.plotting.detector as detect
import wagascianpy.plotting.topology as topol
import wagascianpy.utils.treenames as treenames
import wagascianpy.utils.utils as utils
from wagascianpy.analysis.analysis import SPIROC2D_NCHANNELS, WAGASCI_NCHIPS, WALLMRD_NCHIPS

# import ROOT
try:
    import ROOT

    ROOT.PyConfig.IgnoreCommandLineOptions = True
except ImportError as err:
    if "ROOT" in repr(err):
        ROOT = None
    else:
        raise

# compatible with Python 2 *and* 3:
ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})

# value of the hit TBranch in the raw data TTree when a hit is found
_HIT = 1
# number of columns in a channel
_NCOLUMNS = 16
# number of active channels in a chip
_NCHANNELS = 32
# number of chips in WAGASCI
_NCHIPS_WAGASCI = 20
# number of chips in WallMRD
_CHIPS_WALLMRD = 3


class Patron(object):
    """
    The Patron defines the interface of interest to clients.
    """

    def __init__(self, start=None, stop=None, wagasci_database=None, harvester=None):
        """
        Usually, the Patron accepts a harvester through the constructor, but
        also provides a setter to change it at runtime.
        :param start: start run or start time
        :param stop: stop run or stop time
        :param wagasci_database: wagasci database location
        :param harvester: harvester class
        """
        # type: (Optional[Union[str, int]], Optional[Union[str, int]], Optional[str], Optional[Harvester])
        self._start = start  # type: Optional[Union[str, int]]
        self._stop = stop  # type: Optional[Union[str, int]]
        self._wagasci_database = wagasci_database if wagasci_database else None  # type: Optional[str]
        self._xdata = []
        self._ydata = []
        self._check_arguments()

        self.harvester = harvester

    @property
    def wagasci_database(self):
        # type: (...) -> str
        return self._wagasci_database

    # noinspection PyTypeChecker
    def _check_arguments(self):
        # type: (...) -> None
        """
        Check that the constructor arguments are sane
        """
        if isinstance(self._start, string_types):
            try:
                self._start = wagascianpy.database.db_record.DBRecord.str2datetime(self._start)
            except ValueError as exception:
                print('Start string must be in the format "%Y/%m/%d %H:%M:%S" or a '
                      'run number (int) : ' + self._start)
                raise exception
        if isinstance(self._stop, string_types):
            try:
                self._stop = wagascianpy.database.db_record.DBRecord.str2datetime(self._stop)
            except ValueError as exception:
                print('Stop string must be in the format "%Y/%m/%d %H:%M:%S" or a '
                      'run number (int) : ' + self._stop)
                raise exception

    def is_harvester_ready(self):
        # type: (...) -> bool
        """
        Check if the harvester class has been set
        :return: true if set false otherwise
        """
        return bool(self._harvester)

    @property
    def harvester(self):
        """
        The Patron maintains a reference to one of the Harvester objects. The
        Patron does not know the concrete class of a harvester. It should work
        with all strategies via the Harvester interface.
        """
        assert self._harvester is not None, "Set data harvester before using it"
        return self._harvester

    @harvester.setter
    def harvester(self, harvester):
        """
        Usually, the Patron allows replacing a Harvester object at runtime.
        """
        self._harvester = harvester
        if None not in [self._harvester, self._start, self._wagasci_database]:
            self._harvester.set_time_interval(start=self._start, stop=self._stop,
                                              wagasci_database=self._wagasci_database)

    def gather_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[List[float], List[Any]]
        """
        Call the harvest_data method of the harvester object to gather the data to plot
        :param detector_name: name of the detector if any
        :param only_good: only gather data and runs flagged as good
        :return: two lists of equal length with X axis data and Y axis data
        """
        return self._harvester.harvest_data(detector_name=detector_name, only_good=only_good)


class Harvester(ABC):
    """
    The Harvester interface declares operations common to all supported versions of some algorithm.
    The Patron uses this interface to call the algorithm defined by the concrete Harvester.
    """

    def __init__(self, database, repository, t2krun):
        # type: (str, str, int) -> None
        """
        :param database: location of the BSD or WAGASCI database
        :param repository: location of the BSD or WAGASCI repository
        :param t2krun: number of T2K run
        """
        self._database = database if database else None  # type: Optional[str]
        self._repository = repository if repository else None  # type: Optional[str]
        self._t2krun = t2krun  # type: int
        self._start_time = None
        self._stop_time = None
        self._trees_have_been_planted = False  # type: bool
        self._active_branches = None  # type: Optional[List[str]]
        if self._repository is not None and not os.path.exists(self._repository):
            raise OSError("Repository directory does not exist : %s" % self._repository)
        if self._database is not None and ':' not in self._database and not os.path.exists(self._database):
            raise OSError("Database file does not exist : %s" % self._database)

    def set_time_interval(self, start, stop, wagasci_database=None):
        # type: (Union[str, int], Union[str, int], Optional[str]) -> None
        """
        Set the time or run interval where to look for data
        :param start: start time or start run
        :param stop: stop time or stop run
        :param wagasci_database: location of the WAGASCI database file
        :return: None
        """
        if start is not None:
            database = wagasci_database if wagasci_database is not None else self._database
            self._start_time, self._stop_time = wgdb.run_to_interval(start=start, stop=stop, database=database)

    @abc.abstractmethod
    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[List[float], List[Any]]
        """
        Gather the data to plot
        :param detector_name: name of the detector if any
        :param only_good: only gather data and runs flagged as good
        :return: two lists of equal length with X axis data and Y axis data
        """
        pass

    @staticmethod
    def _list_root_files(run_root_dir, recursive=True):
        # type: (str, bool) -> List[Tuple[str, int]]
        """
        List all the files with .root extension and extract the DIF number
        :param run_root_dir: directory to list
        :return: list of tuples where the first element is the file name and the second is the DIF number
        """
        return [(filename, utils.extract_dif_id(filename))
                for filename in utils.find_files(path=run_root_dir, extension='root', recursive=recursive)
                if utils.extract_dif_id(filename) is not None]


################################################################
#                      Concrete Harvesters                     #
################################################################


class BsdHarvester(Harvester, ABC):

    def _get_spills(self):
        # type: (...) -> List[sp.BsdSpill]
        """
        Read the input files and extract the info about the BSD spills
        :return: list of BsdSpill objects
        """
        return wagascianpy.analysis.beam_summary_data.get_bsd_spills(bsd_database=self._database,
                                                                     bsd_repository=self._repository,
                                                                     t2krun=self._t2krun,
                                                                     start_time=self._start_time,
                                                                     stop_time=self._stop_time)

    @abc.abstractmethod
    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[List[float], List[Any]]
        """
        Gather the data to plot
        :param detector_name: ignored
        :param only_good: only good BSD spills
        :return: two lists of equal length with X axis data and Y axis data
        """
        pass


class BsdPotHarvester(BsdHarvester):

    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[List[float], List[float]]
        """
        Gather the data to plot. X axis is time, Y axis is POT delivered by neutrino beam line
        :param detector_name: ignored
        :param only_good: only good BSD spills
        :return: two lists of equal length with X axis data and Y axis data
        """
        bsd_spills = self._get_spills()
        accumulated_pot_list = []
        accumulated_pot = 0
        timestamp_list = []
        for spill in bsd_spills:
            if not only_good or spill.bsd_good_spill_flag == sp.IS_GOOD_SPILL:
                accumulated_pot += spill.pot
                accumulated_pot_list.append(accumulated_pot)
                timestamp_list.append(spill.timestamp)

        return timestamp_list, accumulated_pot_list


class BsdSpillHarvester(BsdHarvester):

    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[List[float], List[int]]
        """
        Gather the data to plot. X axis is time, Y axis is BSD 32 bit spill number
        :param detector_name: ignored
        :param only_good: only good BSD spills
        :return: two lists of equal length with X axis data and Y axis data
        """
        bsd_spills = self._get_spills()
        spill_number_list = []
        timestamp_list = []
        for spill in bsd_spills:
            if not only_good or spill.bsd_good_spill_flag == sp.IS_GOOD_SPILL:
                spill_number_list.append(spill.bsd_spill_number)
                timestamp_list.append(spill.timestamp)

        return timestamp_list, spill_number_list


class WagasciHarvester(Harvester, ABC):

    def __init__(self, topology=None, *args, **kwargs):
        # type: (Optional[topol.Topology], *Any, **Any) -> None
        super(WagasciHarvester, self).__init__(*args, **kwargs)
        self._detectors = detect.Detectors(enabled_detectors=topology)

    @property
    def active_branches(self):
        # type: (...) -> List[str]
        """
        :return: list of active branches names
        """
        return self._active_branches

    @active_branches.setter
    def active_branches(self, active_branches):
        # type: (List[str]) -> None
        """
        :param active_branches: list of active branches names
        :return: None
        """
        self._active_branches = active_branches

    @staticmethod
    def _is_dif_good(record, dif_id):
        # type: (Dict[str, Any], int) -> bool
        """
        :param record: WAGASCI run record
        :param dif_id: DIF index
        :return: True if the good data flag of the DIF is set to 1, False if it set to 0
        """
        if topol.DifIndex.is_wallmrd_north(dif_id):
            if not record["wallmrd_north_good_data_flag"]:
                return False
        elif topol.DifIndex.is_wallmrd_south(dif_id):
            if not record["wallmrd_south_good_data_flag"]:
                return False
        elif topol.DifIndex.is_wagasci_upstream(dif_id):
            if not record["wagasci_upstream_good_data_flag"]:
                return False
        elif topol.DifIndex.is_wagasci_downstream(dif_id):
            if not record["wagasci_downstream_good_data_flag"]:
                return False
        return True

    def _plant_trees(self, only_good=False):
        # type: (bool) -> None
        """
        Open the input TFiles and assign a TTree object to each enabled DIF object
        :param only_good: only runs or detectors flagged as good
        :return: None
        """
        if not self._trees_have_been_planted:
            # Assign a TTree to each DIF
            with wgdb.WagasciDataBase(db_location=self._database) as db:
                records = db.get_time_interval(datetime_start=self._start_time, datetime_stop=self._stop_time,
                                               include_overlapping=False, only_good=only_good)
                for record in sorted(records, key=operator.itemgetter("run_number")):
                    for root_file, dif_id in self._list_root_files(os.path.join(self._repository, record["name"]),
                                                                   recursive=False):
                        if not only_good or self._is_dif_good(record=record, dif_id=dif_id):
                            tree_name = utils.extract_raw_tree_name(root_file)
                            self._detectors.get_dif(dif_id).add_tree(root_file=root_file, tree_name=tree_name)
            # Set the active branches of the TTree
            self._detectors.activate_branches(self._active_branches)
        self._trees_have_been_planted = True

    def _get_wagasci_spills_from_ttree(self, raw_tree):
        # type: (Any) -> List[sp.WagasciBsdSpill]
        """
        Read all the WAGASCI spills of the input TTree into a list of WagasciBsdSpill objects
        :param raw_tree: ROOT.TTree object
        :return: list of WAGASCI spills
        """
        assert raw_tree is not None, "Raw tree should be set before trying to read it"
        wagasci_spills = []
        for event in raw_tree:
            if event.spill_mode != sp.WAGASCI_SPILL_BEAM_MODE:
                continue
            wagasci_spill = sp.SpillFactory.get_spill("wagascibsd")
            for variable in self._active_branches:
                if not hasattr(event, variable):
                    raise AttributeError("Variable {} not found in TTree {}".format(variable,
                                                                                    raw_tree.GetFile().GetName()))
                setattr(wagasci_spill, variable, getattr(event, variable))
            wagasci_spills.append(wagasci_spill)
        return wagasci_spills

    @abc.abstractmethod
    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> None
        """
        Open the input TTrees and read the spill information from them
        :param detector_name: name of the detector to gather data from
        :param only_good: only gather data and runs flagged as good
        """
        self._plant_trees(only_good)
        assert detector_name is not None, "You must specify a detector where to harvest data from"
        dif = self._detectors.get_detector(detector_name)
        assert isinstance(dif, detect.Dif), "You should select a DIF and not a whole subdetector"
        if dif.has_tree():
            print("Extracting spills from DIF {} of detector {}".format(dif.name, detector_name))
            dif.set_spills(self._get_wagasci_spills_from_ttree(dif.get_tree()))


class WagasciPotHarvester(WagasciHarvester):

    def __init__(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        super(WagasciPotHarvester, self).__init__(*args, **kwargs)
        self._active_branches = ["spill_number", "spill_mode", "fixed_spill_number", "good_spill_flag",
                                 "bsd_good_spill_flag", "pot", "timestamp"]

    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[Optional[List[float]], Optional[List[float]]]
        """
        Return a couple of lists of equal length. The first is the X axis timestamp, the second is the Y axis POTs.
        :param detector_name: name of the detector to gather data from
        :param only_good: only gather data and runs flagged as good
        :return: two lists of equal length with X axis data and Y axis data
        """
        if "top" not in detector_name and "bottom" not in detector_name and "side" not in detector_name:
            detector_name += " top"
        super(WagasciPotHarvester, self).harvest_data(detector_name=detector_name, only_good=only_good)
        top_dif = self._detectors.get_detector(detector_name)
        if top_dif.is_enabled():
            accumulated_pot_list = []
            accumulated_pot = 0
            timestamp_list = []
            for spill in top_dif.get_spills():
                if ((not only_good or spill.bsd_good_spill_flag == sp.IS_GOOD_SPILL) and
                        spill.good_spill_flag == sp.IS_GOOD_SPILL):
                    if spill.timestamp < 0 or spill.pot < 0:
                        print("Huston there was a problem!")
                        spill.pretty_print()
                        continue
                    timestamp_list.append(spill.timestamp)
                    accumulated_pot += spill.pot
                    accumulated_pot_list.append(accumulated_pot)
            return timestamp_list, accumulated_pot_list
        return None, None


class WagasciSpillHistoryHarvester(WagasciHarvester):

    def __init__(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        super(WagasciSpillHistoryHarvester, self).__init__(*args, **kwargs)
        self._active_branches = ["spill_mode", "spill_number", "timestamp"]

    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[Optional[List[float]], Optional[List[int]]]
        """
        Return a couple of lists: the first is the X axis timestamp, the second is the Y axis (non fixed) spill number.
        :param detector_name: name of the detector to gather data from
        :param only_good: only gather data and runs flagged as good
        :return: two lists of equal length with X axis data and Y axis data
        """
        if "top" not in detector_name and "bottom" not in detector_name and "side" not in detector_name:
            detector_name += " top"
        super(WagasciSpillHistoryHarvester, self).harvest_data(detector_name=detector_name, only_good=only_good)
        top_dif = self._detectors.get_detector(detector_name)
        if top_dif.is_enabled():
            spill_number_list = []
            timestamp_list = []
            for spill in top_dif.get_spills():
                if spill.timestamp > 0:
                    spill_number_list.append(spill.spill_number)
                    timestamp_list.append(spill.timestamp)
            return timestamp_list, spill_number_list
        return None, None


class WagasciFixedSpillHarvester(WagasciHarvester):

    def __init__(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        super(WagasciFixedSpillHarvester, self).__init__(*args, **kwargs)
        self._active_branches = ["spill_mode", "fixed_spill_number", "good_spill_flag",
                                 "bsd_good_spill_flag", "timestamp"]

    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[Optional[List[float]], Optional[List[int]]]
        """
        Return a couple of lists: the first is the X axis timestamp, the second is the Y axis fixed spill number.
        :param detector_name: name of the detector to gather data from
        :param only_good: only gather data and runs flagged as good
        :return: two lists of equal length with X axis data and Y axis data
        """
        if "top" not in detector_name and "bottom" not in detector_name and "side" not in detector_name:
            detector_name += " top"
        super(WagasciFixedSpillHarvester, self).harvest_data(detector_name=detector_name, only_good=only_good)
        top_dif = self._detectors.get_detector(detector_name)
        if top_dif.is_enabled():
            fixed_spill_number_list = []
            timestamp_list = []
            for spill in top_dif.get_spills():
                if spill.good_spill_flag == sp.IS_GOOD_SPILL and spill.timestamp > 0:
                    if spill.fixed_spill_number < sp.WAGASCI_MINIMUM_SPILL or \
                            spill.fixed_spill_number > sp.WAGASCI_MAXIMUM_SPILL or \
                            spill.timestamp < 0:
                        print("WARNING! WAGASCI spill is out of range. Time {} : Spill {}".format(
                            spill.timestamp, spill.fixed_spill_number))
                    fixed_spill_number_list.append(spill.fixed_spill_number)
                    timestamp_list.append(spill.timestamp)
            return timestamp_list, fixed_spill_number_list
        return None, None


class WagasciSpillNumberHarvester(WagasciHarvester):

    def __init__(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        super(WagasciSpillNumberHarvester, self).__init__(*args, **kwargs)
        self._active_branches = ["spill_mode", "spill_number"]

    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[Optional[List[int]], Optional[List[int]]]
        """
        Return a couple of lists: the first is the X axis event number, the second is the Y axis (non fixed) spill
        number.
        :param detector_name: name of the detector to gather data from
        :param only_good: only gather data and runs flagged as good
        :return: two lists of equal length with X axis data and Y axis data
        """
        if "top" not in detector_name and "bottom" not in detector_name and "side" not in detector_name:
            detector_name += " top"
        super(WagasciSpillNumberHarvester, self).harvest_data(detector_name=detector_name, only_good=only_good)
        top_dif = self._detectors.get_detector(detector_name)
        if top_dif.is_enabled():
            spill_number_list = []
            event_list = []
            for counter, spill in enumerate(top_dif.get_spills()):
                spill_number_list.append(spill.spill_number)
                event_list.append(counter)
            return event_list, spill_number_list
        return None, None


class TemperatureHarvester(WagasciHarvester):

    def __init__(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        super(TemperatureHarvester, self).__init__(*args, **kwargs)
        self._active_branches = ["spill_mode", "good_spill_flag", "bsd_good_spill_flag",
                                 "timestamp", "temperature"]

    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[Optional[List[float]], Optional[List[float]]]
        """
        Return a couple of lists: the first is the X axis timestamp, the second is the Y axis temperature.
        :param detector_name: name of the detector to gather data from
        :param only_good: only gather data and runs flagged as good
        :return: two lists of equal length with X axis data and Y axis data
        """
        super(TemperatureHarvester, self).harvest_data(detector_name=detector_name, only_good=only_good)
        detector = self._detectors.get_detector(detector_name)
        if detector.is_enabled():
            temperature_list = []
            timestamp_list = []
            for spill in detector.get_spills():
                if ((not only_good or spill.bsd_good_spill_flag == sp.IS_GOOD_SPILL) and
                        spill.good_spill_flag == sp.IS_GOOD_SPILL and spill.timestamp > 0):
                    temperature_list.append(spill.temperature)
                    timestamp_list.append(spill.timestamp)
            return timestamp_list, temperature_list
        return None, None


class HumidityHarvester(WagasciHarvester):

    def __init__(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        super(HumidityHarvester, self).__init__(*args, **kwargs)
        self._active_branches = ["spill_mode", "good_spill_flag", "bsd_good_spill_flag",
                                 "timestamp", "humidity"]

    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[Optional[List[float]], Optional[List[float]]]
        """
        Return a couple of lists: the first is the X axis timestamp, the second is the Y axis humidity.
        :param detector_name: name of the detector to gather data from
        :param only_good: only gather data and runs flagged as good
        :return: two lists of equal length with X axis data and Y axis data
        """
        super(HumidityHarvester, self).harvest_data(detector_name=detector_name, only_good=only_good)
        detector = self._detectors.get_detector(detector_name)
        if detector.is_enabled():
            humidity_list = []
            timestamp_list = []
            for spill in detector.get_spills():
                if ((not only_good or spill.bsd_good_spill_flag == sp.IS_GOOD_SPILL) and
                        spill.good_spill_flag == sp.IS_GOOD_SPILL and spill.timestamp > 0):
                    humidity_list.append(spill.humidity)
                    timestamp_list.append(spill.timestamp)
            return timestamp_list, humidity_list
        return None, None


class AdcCalibrationHarvester(Harvester, ABC):

    def __init__(self, topology=None, *args, **kwargs):
        # type: (topol.Topology, Any, Any) -> None
        """
        :param topology: Topology object that specifies which are the enabled DIFs
        """
        super(AdcCalibrationHarvester, self).__init__(*args, **kwargs)
        self._detectors = detect.Detectors(enabled_detectors=topology)  # type: detect.Detectors

    def _list_adc_calibration_files(self):
        # type: (...) -> List[Tuple[int, int, int, str]]
        rootfiles = utils.find_files(self._repository, extension='root', recursive=False)
        nums_re = re.compile(r"from_(\d+)_to_(\d+)_ecal_dif_(\d+)")
        start_stop_run = collections.namedtuple('start_stop_run', 'start stop dif path')
        results = []
        for path in rootfiles:
            if bool(nums_re.search(path)):
                start = int(nums_re.search(path).group(1))
                stop = int(nums_re.search(path).group(2))
                dif = int(nums_re.search(path).group(3))
                results.append(start_stop_run(start, stop, dif, path))
        return sorted(results)

    @staticmethod
    def _select_run(run_number, start_stop_dirs):
        # type: (int, List[Tuple[int, int, int, str]]) -> List[Tuple[int, int, int, str]]
        return [start_stop_dir for start_stop_dir in start_stop_dirs
                if start_stop_dir.start <= run_number <= start_stop_dir.stop]

    def _plant_trees(self):
        # type: (...) -> None
        """
        Open the input TFiles and assign a TTree object to each enabled DIF object
        :return: None
        """

        if not self._trees_have_been_planted:
            if not self._active_branches:
                raise ValueError("Set active branches before planting trees")

            all_files = self._list_adc_calibration_files()
            selected_files = set()

            if self._database:
                print("[AdbCalibrationHarvester] Using WAGASCI run database to select runs from {} to {}".format(
                    self._start_time, self._stop_time))

                with wgdb.WagasciDataBase(db_location=self._database) as db:
                    records = db.get_time_interval(datetime_start=self._start_time, datetime_stop=self._stop_time,
                                                   include_overlapping=True, only_good=True)
                    for record in sorted(records, key=operator.itemgetter("run_number")):
                        selected_runs = self._select_run(record["run_number"], all_files)
                        for selected_run in selected_runs:
                            selected_files.add(selected_run)
                selected_files = sorted(list(selected_files))
            else:
                print("[AdbCalibrationHarvester] WARNING : WAGASCI run database not given so all runs will be "
                      "selected. The ordering of the runs cannot be guaranteed.")
                selected_files = all_files

            for selected_file in selected_files:
                if self._detectors.get_dif(selected_file.dif).is_enabled():
                    self._detectors.get_dif(selected_file.dif).add_tree(root_file=selected_file.path,
                                                                        tree_name=treenames.HISTORY_TREE_NAME,
                                                                        tree_friends=[treenames.THRESHOLD_TREE_NAME])
            self._detectors.activate_branches(self._active_branches)
        self._trees_have_been_planted = True

    @abc.abstractmethod
    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> None
        """
        Setup the TTrees where to gather data from
        :param detector_name: ignored
        :param only_good: ignored
        :return: None
        """
        self._plant_trees()


class AdcCalibrationHarvesterDif(AdcCalibrationHarvester, ABC):

    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[Optional[List[float]], Optional[List[List[float]]]]
        """
        Return a couple of lists: the first is the timestamp (X axis), the second is the list of gain for all the
        channels (Y axis).
        :param detector_name: name of the detector to gather data from
        :param only_good: ignored
        :return: two lists of equal length with X axis data and Y axis data
        """
        super(AdcCalibrationHarvesterDif, self).harvest_data(detector_name=detector_name, only_good=only_good)
        time = []  # type: List[float]
        data = []  # type: List[List[float]]
        for detector in self._detectors:
            for dif in [dif for dif in detector if dif.has_tree()]:
                print("Extracting data from DIF {} {}".format(detector.name, dif.name))
                for event in dif.get_tree():
                    if getattr(event, self._active_branches[0]) > 0.:
                        yarray = numpy.frombuffer(getattr(event, self._active_branches[1]), dtype="float64")
                        ylist = yarray[~numpy.isnan(yarray)].tolist()
                        ylist = [i for i in ylist if i >= 0.]
                        if ylist:
                            time.append(getattr(event, self._active_branches[0]))
                            data.append(ylist)
        return time, data


class GainHarvesterDif(AdcCalibrationHarvesterDif):

    def __init__(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        super(GainHarvesterDif, self).__init__(*args, **kwargs)
        self._active_branches = ["average_timestamp", "gain"]


class DarkNoiseHarvesterDif(AdcCalibrationHarvesterDif):

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None
        super(DarkNoiseHarvesterDif, self).__init__(*args, **kwargs)
        self._active_branches = ["average_timestamp", "dark_noise"]


class ThresholdHarvesterDif(AdcCalibrationHarvesterDif):

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None
        super(ThresholdHarvesterDif, self).__init__(*args, **kwargs)
        self._active_branches = ["average_timestamp", "hit_threshold"]


class AdcCalibrationTemperatureHarvester(AdcCalibrationHarvester):

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None
        super(AdcCalibrationTemperatureHarvester, self).__init__(*args, **kwargs)
        self._active_branches = ["average_timestamp", "average_temperature"]

    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[Optional[List[float]], Optional[List[float]]]
        """
        Return a couple of lists: the first is the timestamp (X axis), the second is the temperature (Y axis).
        :param detector_name: name of the detector to gather data from
        :param only_good: ignored
        :return: two lists of equal length with X axis data and Y axis data
        """
        super(AdcCalibrationTemperatureHarvester, self).harvest_data(detector_name=detector_name, only_good=only_good)
        time = []
        data = []
        dif = self._detectors.get_dif(detector_name)
        print("Extracting data from DIF {}".format(detector_name))
        for event in dif.get_tree():
            if getattr(event, self._active_branches[0]) > 0. and getattr(event, self._active_branches[1]) > 0.:
                time.append(getattr(event, self._active_branches[0]))
                data.append(getattr(event, self._active_branches[1]))
        return time, data


class AdcCalibrationHarvesterChannel(AdcCalibrationHarvester, ABC):

    def __init__(self, chip=0, channel=0, *args, **kwargs):
        # type: (int, int, Any, Any) -> None
        """
        :param chip: chip ID
        :param channel: channel ID
        """
        super(AdcCalibrationHarvesterChannel, self).__init__(*args, **kwargs)
        self._chip = chip  # type: int
        self._channel = channel  # type: int

    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[Optional[List[float]], Optional[List[float]]]
        """
        Return a couple of lists: the first is the timestamp (X axis), the second is the gain (Y axis).
        :param detector_name: ignored
        :param only_good: ignored
        :return: two lists of equal length with X axis data and Y axis data
        """
        super(AdcCalibrationHarvesterChannel, self).harvest_data(detector_name=detector_name, only_good=only_good)
        time = []
        data = []
        for detector in self._detectors:
            for dif in [dif for dif in detector if dif.has_tree()]:
                print("Extracting data from DIF {} {}".format(detector.name, dif.name))
                for event in dif.get_tree():
                    if getattr(event, self._active_branches[0]) > 0.:
                        yarray = numpy.frombuffer(getattr(event, self._active_branches[1]), dtype="float64")
                        reshaped_array = numpy.reshape(yarray, (WAGASCI_NCHIPS if topol.DifIndex.is_wagasci(dif.dif_id)
                                                                else WALLMRD_NCHIPS, SPIROC2D_NCHANNELS))
                        ydata = reshaped_array[self._chip][self._channel]
                        if ydata >= 0:
                            time.append(getattr(event, self._active_branches[0]))
                            data.append(ydata)
        return time, data


class GainHarvesterChannel(AdcCalibrationHarvesterChannel):

    def __init__(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        super(GainHarvesterChannel, self).__init__(*args, **kwargs)
        self._active_branches = ["average_timestamp", "gain"]


class DarkNoiseHarvesterChannel(AdcCalibrationHarvesterChannel):

    def __init__(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        super(DarkNoiseHarvesterChannel, self).__init__(*args, **kwargs)
        self._active_branches = ["average_timestamp", "dark_noise"]


class ThresholdHarvesterChannel(AdcCalibrationHarvesterChannel):

    def __init__(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        super(ThresholdHarvesterChannel, self).__init__(*args, **kwargs)
        self._active_branches = ["average_timestamp", "hit_threshold"]


class BcidHarvester(WagasciHarvester):

    def __init__(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        """
        :param args: positional arguments for the Harvester super class
        :param kwargs: keyword arguments for the Harvester super class
        """
        super(BcidHarvester, self).__init__(*args, **kwargs)
        self._active_branches = ["spill_mode", "timestamp", "bcid", "hit"]
        # Number of spills over which the BCID histogram is calculated
        self._spill_period = 5000
        # Maximum value of the BCID to be plotted
        self._max_bcid = 100

    def harvest_data(self, detector_name=None, only_good=False):
        # type: (Optional[str], bool) -> Tuple[Optional[List[float]], Optional[List[List[float]]]]
        """
        Setup the TTrees where to gather data from and read the data
        :param detector_name: name of the detector to gather data from
        :param only_good: only gather data and runs flagged as good
        :return: two lists of equal length with X axis data and Y axis data
        """
        super(BcidHarvester, self).harvest_data(detector_name=detector_name, only_good=only_good)
        time = []
        data = []
        for detector in self._detectors:
            for dif in [dif for dif in detector if dif.has_tree() and dif.is_enabled()]:

                if topol.DifIndex.is_wagasci(dif.dif_id):
                    num_chips = _NCHIPS_WAGASCI
                elif topol.DifIndex.is_wallmrd(dif.dif_id):
                    num_chips = _CHIPS_WALLMRD
                else:
                    raise ValueError("DIF ID not recognized : {}".format(dif.dif_id))

                print("Extracting BCID data from DIF {} {}".format(detector.name, dif.name))

                hist = ROOT.TH1I("DIF{}".format(dif.dif_id), "DIF{}".format(dif.dif_id),
                                 self._max_bcid, 0, self._max_bcid)

                average_timestamp = 0
                i = 0
                for event in dif.get_tree():
                    if event.spill_mode == sp.WAGASCI_SPILL_BEAM_MODE and event.timestamp > 0.:
                        i += 1
                        average_timestamp += event.timestamp
                        for chip in range(num_chips):
                            for channel in range(_NCHANNELS):
                                for column in range(_NCOLUMNS):
                                    if event.hit[column + _NCOLUMNS * channel + _NCOLUMNS * _NCHANNELS * chip] == _HIT:
                                        hist.Fill(event.bcid[column + _NCOLUMNS * chip])
                        if i != 0 and i % self._spill_period == 0:
                            print("{} {} : event {}".format(detector.name, dif.name, i))
                            average_timestamp /= self._spill_period
                            yarray = numpy.array(hist)[1:-1]
                            ylist = yarray[~numpy.isnan(yarray)].tolist()
                            ylist = [j for j in ylist if j >= 0.]
                            if ylist:
                                time.append(average_timestamp)
                                data.append(ylist)
                            hist.Reset("ICESM")
                            average_timestamp = 0
        return time, data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio
#

# pylint: disable=broad-except, too-many-instance-attributes, too-many-locals
# pylint: disable=no-self-use, too-many-branches, too-many-statements
# pylint: disable=singleton-comparison, too-many-nested-blocks, arguments-differ

"""WAGASCI simple run database"""
import datetime
import os
import re
import subprocess
from enum import IntEnum
from typing import Optional, List, Union

import wagascianpy.database.db_record
import wagascianpy.database.my_tinydb
import wagascianpy.utils.utils

# compatible with Python 2 *and* 3
try:
    # noinspection PyUnresolvedReferences
    IntTypes = (int, long)  # Python2
except NameError:
    IntTypes = int  # Python3

try:
    import ROOT

    ROOT.PyConfig.IgnoreCommandLineOptions = True
except ImportError as err:
    if "ROOT" in repr(err):
        ROOT = None
    else:
        raise

###############################################################################
#                                  Constants                                  #
###############################################################################

_NON_PHYSICAL_VALUE = -1


class TriggerTime(IntEnum):
    """ Spill time stamp in second for Epoch (UNIX time) """
    GPS1 = 0
    GPS2 = 1
    RubidiumClock = 2


class GoodSpillFlag(IntEnum):
    """
    =0:Bad
    =1:Good@horn250kA
    =-1:Good@horn-250kA
    =2:Good@horn200kA<
    =100:Good@horn0kA)
    """
    Bad = 0
    GoodHornPlus250kA = 1
    GoodHornMinus250kA = -1
    GoodHornLess200kA = 2
    GoodHorn0kA = 100


class NeutrinosType(IntEnum):
    """ Type of neutrinos in the beam """
    Neutrinos = GoodSpillFlag.GoodHornPlus250kA
    AntiNeutrinos = GoodSpillFlag.GoodHornMinus250kA
    Unknown = GoodSpillFlag.Bad


class RunType(IntEnum):
    """ run type (=1:physic run) """
    PhysicsRun = 1


class HornNumber(IntEnum):
    """ Horn number """
    FirstHorn = 0
    SecondHorn = 1
    ThirdHorn = 2


class HornCurrent(IntEnum):
    """ Horn power supply channel """
    TotalCurrent = 0
    Channel1 = 1
    Channel2 = 2
    Channel3 = 3
    Channel4 = 4


class CTNumber(IntEnum):
    """ CT monitor number """
    First = 0
    Second = 1
    Third = 2
    Fourth = 3
    Fifth = 4


class BunchNumber(IntEnum):
    """ Bunch number """
    TotalCurrent = 0
    First = 1
    Second = 2
    Third = 3
    Fourth = 4
    Fifth = 5
    Sixth = 6
    Seventh = 7
    Eighth = 8


###########################################################################
#                                BsdRecord                                #
###########################################################################

class BsdRecord(wagascianpy.database.db_record.DBRecord):
    """Bsd record"""

    def __init__(self, file_path=None, record=None):
        # type: (Optional[str], Optional[BsdRecord]) -> None
        """Initialize BsdRecord from database record"""
        self.file_path = file_path
        self.name = None
        self.t2k_run = None
        self.main_ring_run = None
        self.main_ring_subrun = None
        self.neutrino_daq_run = None
        self.bsd_version = None

        self.start_time = None
        self.stop_time = None
        self.duration_s = None
        self.duration_h = None
        self.start_spill = None
        self.stop_spill = None
        self.number_of_spills = None
        self.number_of_good_spills = None
        self.number_of_pot = None
        self.mean_pot_per_spill = None
        self.mean_horn_current = None
        self.neutrino_type = None

        super(BsdRecord, self).__init__(record)

        if record is None:
            if not isinstance(self.file_path, str):
                raise RuntimeError("Run path is not a string")

            self.name = os.path.basename(self.file_path)

            try:
                match = re.search(r"\S+t2krun(\d+)/bsd_run(\d{3})(\d{4})_(\d{2})(\D\d{2}).root", self.file_path)
                if match is None or len(match.groups()) != 5:
                    raise ValueError("The BSD file path is not valid : %s" % str(self.file_path))
                self.t2k_run = int(match.group(1))
                self.main_ring_run = int(match.group(2))
                self.neutrino_daq_run = int(match.group(2) + match.group(3))
                self.main_ring_subrun = int(match.group(4))
                self.bsd_version = match.group(5)
            except (IndexError, AttributeError) as exception:
                raise RuntimeError("Failed to parse %s : %s" % (str(self.file_path), str(exception)))
            if ROOT:
                try:
                    self._read_tfile()
                except RuntimeError as exception:
                    print(str(exception))
                    self._set_defaults_if_error()
            else:
                self._set_defaults_if_error()

    def _set_defaults_if_error(self):
        # type: (...) -> None
        self.start_time = int(_NON_PHYSICAL_VALUE)
        self.stop_time = int(_NON_PHYSICAL_VALUE)
        self.duration_h = float(_NON_PHYSICAL_VALUE)
        self.duration_s = float(_NON_PHYSICAL_VALUE)
        self.start_spill = int(_NON_PHYSICAL_VALUE)
        self.stop_spill = int(_NON_PHYSICAL_VALUE)
        self.number_of_spills = int(_NON_PHYSICAL_VALUE)
        self.number_of_good_spills = int(_NON_PHYSICAL_VALUE)
        self.number_of_pot = int(_NON_PHYSICAL_VALUE)
        self.mean_pot_per_spill = float(_NON_PHYSICAL_VALUE)
        self.mean_horn_current = float(_NON_PHYSICAL_VALUE)
        self.neutrino_type = NeutrinosType.Unknown

    def _read_tfile(self):
        # type: (...) -> None
        bsd_file = ROOT.TFile.Open(self.file_path)
        if not bsd_file:
            raise RuntimeError("ROOT file %s does not exist or is not valid"
                               % self.file_path)
        bsd_tree = bsd_file.bsd
        if bsd_tree.GetEntries() == 0:
            raise RuntimeError("ROOT file %s is empty" % self.file_path)

        bsd_tree.SetBranchStatus("*", 0)
        bsd_tree.SetBranchStatus("nurun", 1)
        bsd_tree.SetBranchStatus("mrrun", 1)
        bsd_tree.SetBranchStatus("spillnum", 1)
        bsd_tree.SetBranchStatus("trg_sec", 1)
        bsd_tree.SetBranchStatus("hct", 1)
        bsd_tree.SetBranchStatus("good_spill_flag", 1)
        bsd_tree.SetBranchStatus("run_type", 1)

        ct_pot = bsd_tree.GetListOfBranches().FindObject("ct_pot")
        if ct_pot:
            bsd_tree.SetBranchStatus("ct_pot", 1)
        ct_np = bsd_tree.GetListOfBranches().FindObject("ct_np")
        if ct_np:
            bsd_tree.SetBranchStatus("ct_np", 1)

        get_all = 0  # get only active branches
        if bsd_tree.GetEntry(0, get_all) <= 0:
            raise RuntimeError("Failed to read first entry")

        if bsd_tree.nurun != self.neutrino_daq_run:
            raise ValueError("Neutrino DAQ run from TTree ({}) different from one"
                             " from file name ({})".format(bsd_tree.nurun, self.neutrino_daq_run))
        if bsd_tree.mrrun != self.main_ring_run:
            raise ValueError("Main Ring run from TTree ({}) different from one"
                             " from file name ({})".format(bsd_tree.mrrun, self.main_ring_run))

        self.start_spill = bsd_tree.spillnum
        self.start_time = bsd_tree.trg_sec[TriggerTime.RubidiumClock]

        if bsd_tree.GetEntry(bsd_tree.GetEntries() - 1, get_all) <= 0:
            raise RuntimeError("Failed to read last entry")

        self.stop_spill = bsd_tree.spillnum
        self.stop_time = bsd_tree.trg_sec[TriggerTime.RubidiumClock]
        self._set_duration_s()
        self._set_duration_h()

        self.number_of_spills = 0
        self.number_of_good_spills = 0
        self.number_of_pot = 0.0
        self.mean_horn_current = 0.0
        self.mean_pot_per_spill = 0.0
        mean_neutrino_type = 0.0

        for event in bsd_tree:
            self.number_of_spills += 1
            if ((event.good_spill_flag == GoodSpillFlag.GoodHornMinus250kA or
                 event.good_spill_flag == GoodSpillFlag.GoodHornPlus250kA) and
                    event.run_type == RunType.PhysicsRun):
                self.number_of_good_spills += 1
                if ct_pot:
                    self.number_of_pot += event.ct_pot[BunchNumber.TotalCurrent]
                elif ct_np:
                    self.number_of_pot += event.ct_np[CTNumber.Fifth * len(BunchNumber) + BunchNumber.TotalCurrent]
                # noinspection PyTypeChecker
                self.mean_horn_current += event.hct[HornNumber.FirstHorn * len(HornNumber) + HornCurrent.TotalCurrent]
                mean_neutrino_type += event.good_spill_flag
        if self.number_of_good_spills > 0:
            self.mean_horn_current /= self.number_of_good_spills
            self.mean_pot_per_spill = self.number_of_pot / self.number_of_good_spills
            mean_neutrino_type /= self.number_of_good_spills

        if abs(mean_neutrino_type) < 0.1:
            self.neutrino_type = NeutrinosType.Unknown
        elif abs(mean_neutrino_type - float(NeutrinosType.Neutrinos)) < \
                abs(mean_neutrino_type - float(NeutrinosType.AntiNeutrinos)):
            self.neutrino_type = NeutrinosType.Neutrinos
        else:
            self.neutrino_type = NeutrinosType.AntiNeutrinos

    def _set_duration_s(self):
        # type: (...)-> None
        """Set the run duration in seconds
        """
        if self.stop_time is None or self.start_time is None:
            raise ValueError("Set start time and stop time before duration")
        self.duration_s = float((self.stop_time - self.start_time))

    def _set_duration_h(self):
        # type: (...)-> None
        """Set the run duration in hours
        """
        if self.stop_time is None or self.start_time is None:
            raise ValueError("Set start time and stop time before duration")
        self.duration_h = float((self.stop_time - self.start_time) / 3600.0)

    def get_start_datetime(self):
        # type: (...)-> Optional[datetime.datetime]
        """Get start datetime
        """
        if self.start_time is None:
            return None
        return self.timestamp2datetime(self.start_time)

    def get_stop_datetime(self):
        # type: (...)-> Optional[datetime.datetime]
        """Get stop datetime
        """
        if self.stop_time is None:
            return None
        return self.timestamp2datetime(self.stop_time)


###########################################################################
#                               BsdDataBase                               #
###########################################################################

def _check_records(records):
    # type: (List[BsdRecord]) -> None
    """check that all records are sane"""
    for record in records:
        if not isinstance(record, BsdRecord):
            raise ValueError("record is not of type BsdRecord")
        if not record.is_ready():
            raise ValueError("record is not ready to be inserted")


class BsdDataBase(wagascianpy.database.my_tinydb.MyTinyDB):
    """Virtual class to manage a database"""

    def __init__(self,
                 bsd_database_location,  # type: str
                 bsd_repository_location=None,  # type: Optional[str]
                 bsd_repository_download_location=None,  # type: Optional[str]
                 t2kruns=10,  # type: int
                 update_db=False,  # type: bool
                 rebuild_db=False  # type: bool
                 ):
        # type: (...) -> None
        self._repository_location = bsd_repository_location
        if self._repository_location and ':' in self._repository_location:
            self._is_remote_repository = True
        else:
            self._is_remote_repository = False
        self._repository_download_location = bsd_repository_download_location
        self._t2kruns = [t2kruns] if isinstance(t2kruns, int) else t2kruns
        super(BsdDataBase, self).__init__(bsd_database_location)
        if rebuild_db:
            self.clear_database()
            self._is_fresh_database = True
        if update_db or rebuild_db:
            self._update_bsd_db()

    def _rsync(self):
        # type: (...) -> None
        rsync = wagascianpy.utils.utils.which("rsync")
        if rsync is None:
            raise RuntimeError("rsync program not found")
        remote = ""
        local = ""
        try:
            for t2krun in self._t2kruns:
                remote = "{}/t2krun{}/*".format(self._repository_location, t2krun)
                local = "{}/t2krun{}/".format(self._repository_download_location, t2krun)
                if self._is_remote_repository:
                    subprocess.check_output([rsync, "-a", "-essh", remote, local])
                else:
                    subprocess.check_output([rsync, "-a", remote, local])
        except subprocess.CalledProcessError as exception:
            raise RuntimeError("Error while copying the remote repository %s into the "
                               "local directory %s : %s" % (remote, local, str(exception)))

    def _update_bsd_db(self):
        # type: (...) -> None
        if self._is_remote_repository:
            if not self._repository_download_location:
                raise ValueError("If updating the remote repository you must specify a download path")
            else:
                download_location = self._repository_download_location
        else:
            download_location = self._repository_location
        with wagascianpy.utils.utils.Cd(download_location):
            if download_location != self._repository_location:
                # Copy all the remote BSD files into a local folder
                self._rsync()
            # List all the BSD files
            run_list = []
            for root, _, files in os.walk(download_location):
                for afile in files:
                    if '.root' in afile:
                        run_list.append(os.path.join(root, afile))

            run_record_list = []
            # Loop over every BSD file
            for run_path in sorted(run_list):
                if self.has_record(os.path.basename(run_path)):
                    print("Skipping run %s" % run_path)
                    continue
                else:
                    print("Run %s not found in database" % run_path)
                    run_record_list.append(BsdRecord(file_path=run_path))

        _check_records(run_record_list)
        self.update_database(run_record_list)

    def get_time_interval(self,
                          datetime_start,  # type: Union[datetime, str, IntTypes, float]
                          datetime_stop,  # type: Union[datetime, str, IntTypes, float]
                          *args,
                          **kwargs):
        # type: (...) -> List[BsdRecord]
        """ Get BSD records in an interval of time """
        if self.is_fresh_database():
            raise RuntimeWarning("BSD database is empty")
        timestamp_start = BsdRecord.datetime2timestamp(datetime_start)
        timestamp_stop = BsdRecord.datetime2timestamp(datetime_stop)
        documents = super(BsdDataBase, self).get_time_interval(timestamp_start, timestamp_stop, *args, **kwargs)
        records = []
        for document in documents:
            records.append(BsdRecord(record=document))
        return records

    def pretty_print(self, name):
        # type: (str) -> None
        """ Print info about a run """
        BsdRecord(record=self.get_record(name)[0]).pretty_print()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio
#

from collections import namedtuple

from typing import List

from wagascianpy.viewer.import_tkinter import import_pygubu

pygubu = import_pygubu()

DetectorState = namedtuple('_State', ['index', 'enabled'])


class Topology(object):
    """List of enabled and disabled detectors"""

    def __init__(self):
        self.WallmrdNorthTop = DetectorState(0, True)
        self.WallmrdNorthBottom = DetectorState(1, True)
        self.WallmrdSouthTop = DetectorState(2, True)
        self.WallmrdSouthBottom = DetectorState(3, True)
        self.WagasciUpstreamTop = DetectorState(4, True)
        self.WagasciUpstreamSide = DetectorState(5, True)
        self.WagasciDownstreamTop = DetectorState(6, True)
        self.WagasciDownstreamSide = DetectorState(7, True)

    def are_all_enabled(self):
        # type: (...) -> bool
        for value in vars(self).values():
            if value.enabled is False:
                return False
        return True

    def get_all(self):
        # type: (...) -> List[DetectorState]
        return [value for value in vars(self).values()]

    def get_enabled(self):
        # type: (...) -> List[DetectorState]
        return [value for value in vars(self).values() if value.enabled is True]

    def get_disabled(self):
        # type: (...) -> List[DetectorState]
        return [value for value in vars(self).values() if value.enabled is False]

    # noinspection PyProtectedMember
    def read_topology(self, builder):
        # type: (pygubu.Builder) -> None
        self.WallmrdNorthTop = self.WallmrdNorthTop._replace(
            enabled=builder.tkvariables.__getitem__('wallmrd_top_north').get())
        self.WallmrdNorthBottom = self.WallmrdNorthBottom._replace(
            enabled=builder.tkvariables.__getitem__('wallmrd_bottom_north').get())
        self.WallmrdSouthTop = self.WallmrdSouthTop._replace(
            enabled=builder.tkvariables.__getitem__('wallmrd_top_south').get())
        self.WallmrdSouthBottom = self.WallmrdSouthBottom._replace(
            enabled=builder.tkvariables.__getitem__('wallmrd_bottom_south').get())
        self.WagasciUpstreamTop = self.WagasciUpstreamTop._replace(
            enabled=builder.tkvariables.__getitem__('wagasci_top_upstream').get())
        self.WagasciUpstreamSide = self.WagasciUpstreamSide._replace(
            enabled=builder.tkvariables.__getitem__('wagasci_side_upstream').get())
        self.WagasciDownstreamTop = self.WagasciDownstreamTop._replace(
            enabled=builder.tkvariables.__getitem__('wagasci_top_downstream').get())
        self.WagasciDownstreamSide = self.WagasciDownstreamSide._replace(
            enabled=builder.tkvariables.__getitem__('wagasci_side_downstream').get())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

import math
from collections import namedtuple

import ROOT
import numpy
from six import string_types
from typing import List, Optional, Union

import wagascianpy.plotting.colors

ROOT.PyConfig.IgnoreCommandLineOptions = True

# compatible with Python 2 *and* 3
try:
    # noinspection PyUnresolvedReferences
    IntTypes = (int, long)  # Python2
except NameError:
    IntTypes = int  # Python3

Range = namedtuple('Range', ['lower_bound', 'upper_bound'])


class Graph(object):
    _num_logscale_bins = 100

    def __init__(self, title, graph_id=None):
        # type: (str, Optional[str]) -> None
        self.title = title
        self.id = graph_id if graph_id is not None else title
        self._xdata = None
        self._ydata = None
        self._color = wagascianpy.plotting.colors.Colors.Black.value
        self._yaxis_color = wagascianpy.plotting.colors.Colors.Black.value
        self._xrange = Range(lower_bound=None, upper_bound=None)
        self._yrange = Range(lower_bound=None, upper_bound=None)
        self._is_empty = True
        self._raw_ydata = True

    @property
    def xdata(self):
        # type: (...) -> Optional[List]
        return self._xdata

    @xdata.setter
    def xdata(self, xdata):
        # type: (List) -> None
        if isinstance(xdata, list):
            self._xdata = numpy.asarray(xdata, dtype=numpy.float64)
        elif isinstance(xdata, numpy.ndarray):
            self._xdata = xdata
        elif xdata is None:
            self._xdata = numpy.array([], dtype=numpy.float64)
        else:
            raise TypeError("Data format not recognized : type(ydata) = {}".format(type(xdata).__name__))
        assert self._xdata.dtype == numpy.float64
        if self._xdata.size == 0:
            self._is_empty = True
        else:
            self._is_empty = False

    @property
    def ydata(self):
        # type: (...) -> Optional[List]
        return self._ydata

    @ydata.setter
    def ydata(self, ydata):
        # type: (List) -> None
        if isinstance(ydata, list):
            self._ydata = numpy.asarray(ydata, dtype=numpy.float64)
        elif isinstance(ydata, numpy.ndarray):
            self._ydata = ydata
        elif ydata is None:
            self._ydata = numpy.array([], dtype=numpy.float64)
        else:
            raise TypeError("Data format not recognized : type(ydata) = {}".format(type(ydata).__name__))
        assert self._ydata.dtype == numpy.float64
        if self._ydata.size == 0:
            self._is_empty = True
        else:
            self._is_empty = False

    @staticmethod
    def _parse_color(color):
        # type: (Union[int, wagascianpy.plotting.colors.Colors, str]) -> int
        if isinstance(color, IntTypes):
            output = color
        elif isinstance(color, wagascianpy.plotting.colors.Colors):
            output = color.value
        elif isinstance(color, string_types):
            output = wagascianpy.plotting.colors.Colors.get_by_detector(color).value
        else:
            raise TypeError("Color type is unknown {}".format(type(color).__name__))
        return output

    @property
    def color(self):
        # type: (...) -> Optional[List]
        return self._color

    @color.setter
    def color(self, color):
        # type: (Union[int, wagascianpy.plotting.colors.Colors, str]) -> None
        self._color = self._parse_color(color)

    @property
    def yaxis_color(self):
        return self._yaxis_color

    @yaxis_color.setter
    def yaxis_color(self, color):
        # type: (Union[int, wagascianpy.plotting.colors.Colors, str]) -> None
        self._yaxis_color = self._parse_color(color)

    @property
    def xrange(self):
        # type: (...) -> Optional[Range]
        return self._xrange

    @xrange.setter
    def xrange(self, myrange):
        # type: (Range) -> None
        if isinstance(myrange, Range):
            self._xrange = myrange
        else:
            raise TypeError("X axis Range type is unknown {}".format(type(myrange).__name__))

    @property
    def yrange(self):
        # type: (...) -> Optional[Range]
        return self._yrange

    @yrange.setter
    def yrange(self, yrange):
        # type: (Range) -> None
        if isinstance(yrange, Range):
            self._yrange = yrange
        else:
            raise TypeError("Y axis Range type is unknown {}".format(type(yrange).__name__))

    def is_empty(self):
        # type: (...) -> bool
        return self._is_empty

    def make_tgraph(self, logscale=False):
        # type: (...) -> ROOT.TGraph
        if self._xdata is None or self._ydata is None:
            raise RuntimeError("Please insert some data in the graph")
        if len(self._xdata) != len(self._ydata):
            raise IndexError("xdata ({}) and ydata ({}) length mismatch".format(len(self._xdata), len(self._ydata)))
        tgraph = ROOT.TGraph(len(self._xdata), self._xdata, self._ydata)
        tgraph.SetNameTitle(self.id, self.title)
        tgraph.SetLineColor(self._color)
        tgraph.SetMarkerColor(self._color)
        x_lower_bound = self._xrange.lower_bound if self._xrange.lower_bound is not None else \
            numpy.amin(self._xdata)
        x_upper_bound = self._xrange.upper_bound if self._xrange.upper_bound is not None else \
            numpy.amax(self._xdata)
        y_lower_bound = self._yrange.lower_bound if self._yrange.lower_bound is not None else \
            numpy.amin(self._ydata)
        y_upper_bound = self._yrange.upper_bound if self._yrange.upper_bound is not None else \
            numpy.amax(self._ydata)
        tgraph.GetXaxis().SetLimits(x_lower_bound, x_upper_bound)
        tgraph.GetHistogram().SetMaximum(y_upper_bound)
        tgraph.GetHistogram().SetMinimum(y_lower_bound)
        tgraph.GetYaxis().SetTitleColor(self.yaxis_color)
        tgraph.GetYaxis().SetLabelColor(self.yaxis_color)
        tgraph.GetYaxis().SetAxisColor(self.yaxis_color)
        return tgraph


class Graph2D(Graph):

    @property
    def raw_ydata(self):
        # type: (...) -> bool
        return self._raw_ydata

    @raw_ydata.setter
    def raw_ydata(self, raw_ydata):
        # type: (bool) -> None
        self._raw_ydata = raw_ydata

    @property
    def ydata(self):
        # type: (...) -> Optional[List]
        return self._ydata

    @ydata.setter
    def ydata(self, ydata):
        # type: (Optional[List[List[float]]]) -> None
        if ydata is None:
            self._ydata = []
            self._is_empty = True
        elif not isinstance(ydata, list):
            self._ydata = []
            self._is_empty = True
            raise TypeError("Data format not recognized : type(ydata) = {}".format(type(ydata).__name__))
        elif len(ydata) == 0:
            self._ydata = []
            self._is_empty = True
        else:
            self._ydata = ydata
            self._is_empty = False

    def _create_logscale_bins(self, min_bin, max_bin):
        # type: (float, float) -> numpy.ndarray
        log_min = math.log10(min_bin)
        log_max = math.log10(max_bin)
        num_bins = self._num_logscale_bins
        dlog = (log_max - log_min) / float(num_bins)
        bins = []
        for i in range(num_bins + 1):
            ibin = log_min + i * dlog
            bins.append(math.pow(10, ibin))
        return numpy.array(bins, dtype=numpy.float64)

    def make_tgraph(self, logscale=False):
        # type: (...) -> ROOT.TH2D
        if self.xdata is None or self.ydata is None:
            raise RuntimeError("Please insert some data in the graph")
        if len(self._xdata) != len(self.ydata):
            raise IndexError("xdata ({}) and ydata ({}) length mismatch".format(len(self.xdata), len(self.ydata)))
        num_history_bins = len(self.xdata)
        min_history_bin = min(self.xdata)
        max_history_bin = max(self.xdata)
        if self._raw_ydata:
            flat_ydata = [item for sublist in self.ydata for item in sublist]
            min_ybin = int(min(flat_ydata))
            max_ybin = int(max(flat_ydata))
        else:
            min_ybin = 0
            max_ybin = len(self.ydata[0])
        if logscale:
            min_ybin -= int(0.1 * abs(max_ybin - min_ybin))
            min_ybin = max(1., min_ybin)
            max_ybin += int(0.1 * abs(max_ybin - min_ybin))
            th2d = ROOT.TH2D(self.id, self.title,
                             num_history_bins, min_history_bin, max_history_bin,
                             self._num_logscale_bins, self._create_logscale_bins(min_ybin, max_ybin))
        else:
            if self._raw_ydata:
                min_ybin -= int(0.1 * abs(max_ybin - min_ybin))
                max_ybin += int(0.1 * abs(max_ybin - min_ybin))
            num_ybins = abs(max_ybin - min_ybin)
            th2d = ROOT.TH2D(self.id, self.title,
                             num_history_bins, min_history_bin, max_history_bin,
                             num_ybins, min_ybin, max_ybin)
        for xbin, xdata in enumerate(self.xdata):
            if self._raw_ydata:
                for ydata in self.ydata[xbin]:
                    th2d.Fill(xdata, ydata)
            else:
                for ybin, ydata in enumerate(self.ydata[xbin]):
                    th2d.SetBinContent(xbin + 1, ybin + 1, ydata)
        ROOT.gStyle.SetPalette(ROOT.kRainBow)
        ROOT.gStyle.SetOptStat("MR")
        ROOT.gStyle.SetNumberContours(100)
        th2d.SetMarkerStyle(ROOT.kFullDotMedium)
        th2d.GetXaxis().SetNdivisions(9, 3, 0, ROOT.kTRUE)
        th2d.GetXaxis().SetTimeDisplay(1)
        th2d.GetXaxis().SetTimeFormat("#splitline{%d %b}{%H:%M}%F1970-01-01 00:00:00")
        th2d.GetXaxis().SetLabelOffset(0.03)
        th2d.SetStats(ROOT.kFALSE)
        th2d.GetYaxis().SetTitleColor(self.yaxis_color)
        th2d.GetYaxis().SetLabelColor(self.yaxis_color)
        th2d.GetYaxis().SetAxisColor(self.yaxis_color)
        return th2d



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# copyright 2020 pintaudi giorgio
#

import argparse
import sys
import time

import wagascianpy.plotting.topology as topol
from wagascianpy.analysis.analysis import WAGASCI_NCHIPS, WALLMRD_NCHIPS
from wagascianpy.utils.configuration import *


def _check_topology_sanity(dif, chip, channel):
    elements = [dif, chip, channel]
    if any(i < 0 for i in elements) and not all(i >= 0 for i in elements):
        raise ValueError("Or all of dif, chip, channel are >= 0, or no one should be!")
    if topol.DifIndex.is_wagasci(dif):
        if chip >= WAGASCI_NCHIPS:
            raise ValueError("WAGASCI chip ID {} is out of boundary {}".format(dif, WAGASCI_NCHIPS))
    elif topol.DifIndex.is_wallmrd(dif):
        if chip >= WALLMRD_NCHIPS:
            raise ValueError("WallMRD chip ID {} is out of boundary {}".format(dif, WALLMRD_NCHIPS))
    else:
        ValueError("DIF ID {} not recognized".format(dif))


def fill_configuration(args=None):
    # type: (Optional[[argparse.Namespace, Dict[str, Any]]]) -> None
    """
    Take as input the parsed arguments from command line and return the Configuration object
    :param args: parsed arguments
    :return: Configuration object
    """

    if isinstance(args, argparse.Namespace):
        args = vars(args)

    # plotting configuration
    output_string = args["output_string"]
    output_path = args["output_path"]
    delivered_pot = args["delivered_pot"]
    accumulated_pot = args["accumulated_pot"]
    bsd_spill_history = args["bsd_spill_history"]
    wagasci_spill_history = args["wagasci_spill_history"]
    wagasci_spill_number = args["wagasci_spill_number"]
    wagasci_fixed_spill = args["wagasci_fixed_spill"]
    temperature = args["temperature"]
    humidity = args["humidity"]
    gain_history = args["gain_history"]
    dark_noise_history = args["dark_noise_history"]
    threshold_history = args["threshold_history"]
    bcid_history = args["bcid_history"]
    all_plotters = args["all"]
    run_markers = args["run_markers"]
    maintenance_markers = args["maintenance_markers"]
    trouble_markers = args["trouble_markers"]
    bsd_markers = args["bsd_markers"]
    topology = args["topology"]
    only_good = args["only_good"]
    save_tfile = args["save_tfile"]
    randomize_channels = args["randomize_channels"]
    dif = args["dif"]
    chip = args["chip"]
    channel = args["channel"]
    # WAGASCI database configuration
    wagasci_database = args["wagasci_database"]
    wagasci_decoded_location = args["wagasci_decoded_location"]
    # BSD database configuration
    bsd_database = args["bsd_database"]
    bsd_repository = args["bsd_repository"]
    bsd_download_location = args["bsd_download_location"]
    # global configuration
    t2krun = args["t2krun"]
    history_location = args["history_location"]
    # run selectors
    start_time = args["start_time"]
    stop_time = args["stop_time"]
    start_run = args["start_run"]
    stop_run = args["stop_run"]

    # wagasci database configuration
    if not wagasci_database:
        wagasci_database = Configuration.wagascidb.wagasci_database()
    if not wagasci_decoded_location:
        wagasci_decoded_location = Configuration.wagascidb.wagasci_decoded_location()

    # bsd database configuration
    if not bsd_database:
        bsd_database = Configuration.bsddb.bsd_database()
    if not bsd_download_location:
        bsd_download_location = Configuration.bsddb.bsd_download_location()
    if not bsd_repository:
        bsd_repository = Configuration.bsddb.bsd_repository()

    # global configuration
    if not t2krun:
        t2krun = Configuration.global_configuration.t2krun()
    if not history_location:
        history_location = Configuration.global_configuration.history_location()

    # sanity checks
    if all_plotters:
        delivered_pot = accumulated_pot = bsd_spill_history = wagasci_spill_history = True
        wagasci_spill_number = wagasci_fixed_spill = temperature = humidity = True
        gain_history = dark_noise_history = threshold_history = True

    if randomize_channels > 0:
        dif = chip = channel = -1
    elif dif >= 0 or chip >= 0 or channel >= 0:
        _check_topology_sanity(dif=dif, chip=chip, channel=channel)

    start = start_run if start_run else start_time
    stop = stop_run if stop_run else stop_time

    if start and stop and not wagasci_database:
        raise KeyError("WAGASCI database must also be specified if start and stop arguments are used")

    # WAGASCI database configuration
    Configuration.wagascidb.override({
        'wagasci_database': wagasci_database,
        'wagasci_decoded_location': wagasci_decoded_location
    })

    # BSD database configuration
    Configuration.bsddb.override({
        'bsd_database': bsd_database,
        'bsd_download_location': bsd_download_location,
        'bsd_repository': bsd_repository
    })

    # global configuration
    Configuration.global_configuration.override({
        't2krun': t2krun,
        'history_location': history_location
    })

    # plotter configuration
    Configuration.create_section('plotter')
    Configuration.plotter.override({
        'output_string': output_string,
        'output_path': output_path,
        'delivered_pot': delivered_pot,
        'accumulated_pot': accumulated_pot,
        'bsd_spill_history': bsd_spill_history,
        'wagasci_spill_history': wagasci_spill_history,
        'wagasci_spill_number': wagasci_spill_number,
        'wagasci_fixed_spill': wagasci_fixed_spill,
        'temperature': temperature,
        'humidity': humidity,
        'gain_history': gain_history,
        'dark_noise_history': dark_noise_history,
        'threshold_history': threshold_history,
        'bcid_history': bcid_history,
        'run_markers': run_markers,
        'maintenance_markers': maintenance_markers,
        'trouble_markers': trouble_markers,
        'bsd_markers': bsd_markers,
        'save_tfile': save_tfile,
        'topology': topology,
        'randomize_channels': randomize_channels,
        'dif': dif,
        'chip': chip,
        'channel': channel,
        'only_good': only_good
    })

    # run selectors
    Configuration.create_section('run_select')
    Configuration.run_select.override({
        'start': start,
        'stop': stop,
    })

    print("CONFIGURATION:")
    Configuration.dump()
    sys.stdout.flush()
    time.sleep(1)



import abc
from collections import namedtuple

import numpy

# compatible with Python 2 *and* 3
from typing import Union, Type, List

# ROOT
try:
    import ROOT
    ROOT.PyConfig.IgnoreCommandLineOptions = True
except ImportError:
    ROOT = None

import wagascianpy.utils

try:
    # noinspection PyUnresolvedReferences
    IntTypes = (int, long)  # Python2
except NameError:
    IntTypes = int  # Python3

# compatible with Python 2 *and* 3:
ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})

_NON_INITIALIZED_VALUE = -2

WAGASCI_SPILL_BEAM_MODE = 1
WAGASCI_SPILL_NO_BEAM_MODE = 0
IS_GOOD_SPILL = 1
NOT_GOOD_SPILL = 0
WAGASCI_MINIMUM_SPILL = 32768
WAGASCI_MAXIMUM_SPILL = 65535

SpillArrayInfo = namedtuple('SpillArrayInfo', ['name', 'array', 'type_char', 'type_str'])


class VirtualSpill(ABC):

    def are_all_defined(self):
        # type: (...) -> bool
        """
        Check if all fields of the spill object are defined (not undefined)
        :return: true if all fields are defined, false otherwise
        """
        for key, value in self.__dict__.items():
            if isinstance(value, (IntTypes, float, bool)):
                assert (value != _NON_INITIALIZED_VALUE), "instance variable {} is still not initialized".format(key)
            elif isinstance(value, numpy.ndarray):
                assert (not numpy.allclose(value, numpy.full(8, fill_value=_NON_INITIALIZED_VALUE,
                                                             dtype=numpy.float64))), \
                    "instance variable {} is still not initialized".format(key)
        return True

    def pretty_print(self):
        """
        Print information about the spill
        :return: None
        """
        for key, value in self.__dict__.items():
            print("%s = %s" % (key, value))

    def _get_type_char(self, var_name):
        # type: (str) -> str
        """
        Get the character corresponding to the type of an attribute of the object.
        Only float, int and bool types are supported
        :param var_name: name of the attribute
        :return: type character
        """
        assert (hasattr(self, var_name)), \
            "object {} has not an attribute named {}".format(type(self).__name__, var_name)
        var = getattr(self, var_name)
        assert (isinstance(var, (IntTypes, float, bool, numpy.ndarray))), \
            "type of variable {} is not supported : {}".format(var_name, type(var_name).__name__)
        if isinstance(var, numpy.ndarray):
            var = var[0]
        if isinstance(var, bool):
            type_char = 'o'
        elif isinstance(var, IntTypes):
            type_char = 'i'
        elif isinstance(var, float):
            type_char = 'd'
        else:
            raise NotImplementedError("Type not recognized")
        return type_char

    def _get_numpy_type(self, var_name):
        # type: (str) -> Type[Union[numpy.int32, numpy.float64, numpy.bool]]
        """
        Get the numpy type corresponding to the type of an attribute of the object.
                Only float, int and bool types are supported
        :param var_name: name of the attribute
        :return: numpy type
        """
        type_char = self._get_type_char(var_name)
        if type_char == 'i':
            type_string = numpy.int32
        elif type_char == 'd':
            type_string = numpy.float64
        elif type_char == 'o':
            type_string = numpy.bool
        else:
            raise NotImplementedError("Type character not recognized : %s" % type_char)
        return type_string

    def _get_type_str(self, var_name):
        # type: (str) -> str
        """
        Get a string corresponding to the type of an attribute of the object.
        Only float, int and bool types are supported. This string is going to
        be passed to a TTree when creating a TBranch.
        :param var_name: name of the attribute
        :return: numpy type
        """
        type_char = self._get_type_char(var_name)
        value = getattr(self, var_name)
        if isinstance(value, numpy.ndarray) and value.size > 1:
            return '{}[{}]/{}'.format(var_name.strip('_'), value.size, type_char.upper())
        else:
            return '{}/{}'.format(var_name.strip('_'), type_char.upper())

    def _get_array(self, var_name):
        # type: (str) -> numpy.array
        """
        Create a numpy array containing an attribute of the object.
        :param var_name: name of the attribute
        :return: numpy array
        """
        assert (hasattr(self, var_name)), \
            "object {} has not an attribute named {}".format(type(self).__name__, var_name)
        value = getattr(self, var_name)
        if not isinstance(value, numpy.ndarray):
            return numpy.array([value], dtype=self._get_numpy_type(var_name))
        else:
            return value

    def _set_array(self, array_info):
        # type: (SpillArrayInfo) -> None
        """
        Set the array field of the SpillArrayInfo tuple to the relative attribute of this object
        :param array_info: SpillArrayInfo tuple
        :return: None
        """
        assert (hasattr(self, array_info.name)), \
            "object {} has not an attribute named {}".format(type(self).__name__, array_info.name)
        value = getattr(self, array_info.name)
        if isinstance(value, numpy.ndarray):
            numpy.copyto(array_info.array, value)
        else:
            array_info.array[0] = getattr(self, array_info.name)

    def get_array_list(self):
        # type: (...) -> List[SpillArrayInfo]
        """
        Create a list of SpillArrayInfo where each element contains an attribute of the object
        :return: list of SpillArrayInfo
        """
        array_list = []
        for key in self.__dict__:
            array_info = SpillArrayInfo(name=key.strip('_'),
                                        array=self._get_array(key),
                                        type_char=self._get_type_char(key),
                                        type_str=self._get_type_str(key))
            array_list.append(array_info)
        return array_list

    def set_array_list(self, array_list):
        # type: (List[SpillArrayInfo]) -> None
        """
        Set each element of a list of SpillArrayInfo to the relative attribute of the object
        :param array_list:
        :return:
        """
        for array_info in array_list:
            self._set_array(array_info)


class BsdSpill(VirtualSpill):
    """
    BSD spill
    """

    def __init__(self):
        self._bsd_spill_number = int(_NON_INITIALIZED_VALUE)
        self._converted_spill_number = int(_NON_INITIALIZED_VALUE)
        self._pot = float(_NON_INITIALIZED_VALUE)
        self._timestamp = float(_NON_INITIALIZED_VALUE)
        self._bsd_good_spill_flag = int(_NON_INITIALIZED_VALUE)
        self._t2k_run = int(_NON_INITIALIZED_VALUE)
        self._main_ring_run = int(_NON_INITIALIZED_VALUE)
        self._neutrino_daq_run = int(_NON_INITIALIZED_VALUE)
        self._horn_current = float(_NON_INITIALIZED_VALUE)
        self._neutrino_mode = int(_NON_INITIALIZED_VALUE)
        self._bunch_pot = numpy.full(8, fill_value=_NON_INITIALIZED_VALUE, dtype=numpy.float64)

    @property
    def bsd_spill_number(self):
        return self._bsd_spill_number

    @bsd_spill_number.setter
    def bsd_spill_number(self, bsd_spill_number):
        self._bsd_spill_number = int(bsd_spill_number)

    @property
    def converted_spill_number(self):
        return self._converted_spill_number

    @converted_spill_number.setter
    def converted_spill_number(self, converted_spill_number):
        self._converted_spill_number = int(converted_spill_number)

    @property
    def pot(self):
        return self._pot

    @pot.setter
    def pot(self, pot):
        self._pot = float(pot)

    @property
    def t2k_run(self):
        return self._t2k_run

    @t2k_run.setter
    def t2k_run(self, t2k_run):
        self._t2k_run = int(t2k_run)

    @property
    def main_ring_run(self):
        return self._main_ring_run

    @main_ring_run.setter
    def main_ring_run(self, main_ring_run):
        self._main_ring_run = int(main_ring_run)

    @property
    def neutrino_daq_run(self):
        return self._neutrino_daq_run

    @neutrino_daq_run.setter
    def neutrino_daq_run(self, neutrino_daq_run):
        self._neutrino_daq_run = int(neutrino_daq_run)

    @property
    def neutrino_mode(self):
        return self._neutrino_mode

    @neutrino_mode.setter
    def neutrino_mode(self, neutrino_mode):
        self._neutrino_mode = int(neutrino_mode)

    @property
    def horn_current(self):
        return self._horn_current

    @horn_current.setter
    def horn_current(self, horn_current):
        self._horn_current = float(horn_current)

    @property
    def bsd_good_spill_flag(self):
        return self._bsd_good_spill_flag

    @bsd_good_spill_flag.setter
    def bsd_good_spill_flag(self, bsd_good_spill_flag):
        self._bsd_good_spill_flag = int(bsd_good_spill_flag)

    @property
    def bunch_pot(self):
        return self._bunch_pot

    @bunch_pot.setter
    def bunch_pot(self, bunch_pot):
        if isinstance(bunch_pot, list):
            self._bunch_pot = numpy.array(bunch_pot, dtype=numpy.float64)
        elif isinstance(bunch_pot, numpy.ndarray):
            self._bunch_pot = bunch_pot
        else:
            raise TypeError("invalid bunch_pot type : {}".format(type(bunch_pot)))

    @property
    def timestamp(self):
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp):
        self._timestamp = float(timestamp)


class WagasciSpill(VirtualSpill):
    """
    WAGASCI spill number
    """

    def __init__(self):
        self._spill_count = int(_NON_INITIALIZED_VALUE)
        self._spill_number = int(_NON_INITIALIZED_VALUE)
        self._converted_spill_number = int(_NON_INITIALIZED_VALUE)
        self._spill_mode = int(_NON_INITIALIZED_VALUE)
        self._good_spill_flag = bool(_NON_INITIALIZED_VALUE)
        self._temperature = 0.
        self._humidity = 0.
        self._wagasci_run = int(_NON_INITIALIZED_VALUE)

    @property
    def spill_count(self):
        return self._spill_count

    @spill_count.setter
    def spill_count(self, spill_count):
        self._spill_count = int(spill_count)

    @property
    def spill_number(self):
        return self._spill_number

    @spill_number.setter
    def spill_number(self, spill_number):
        self._spill_number = int(spill_number)

    @property
    def spill_mode(self):
        return self._spill_mode

    @spill_mode.setter
    def spill_mode(self, spill_mode):
        self._spill_mode = int(spill_mode)

    @property
    def converted_spill_number(self):
        return self._converted_spill_number

    @converted_spill_number.setter
    def converted_spill_number(self, converted_spill_number):
        self._converted_spill_number = int(converted_spill_number)

    @property
    def good_spill_flag(self):
        return self._good_spill_flag

    @good_spill_flag.setter
    def good_spill_flag(self, good_spill_flag):
        assert good_spill_flag in [0, 1, False, True, "0", "1", "False", "True", "false", "true"], \
            "Invalid value of good spill flag"
        self._good_spill_flag = bool(good_spill_flag)

    @property
    def temperature(self):
        return self._temperature

    @temperature.setter
    def temperature(self, temperature):
        self._temperature = float(temperature)

    @property
    def humidity(self):
        return self._humidity

    @humidity.setter
    def humidity(self, humidity):
        self._humidity = float(humidity)

    @property
    def wagasci_run(self):
        return self._wagasci_run

    @wagasci_run.setter
    def wagasci_run(self, wagasci_run):
        self._wagasci_run = int(wagasci_run)


class WagasciBsdSpill(WagasciSpill, BsdSpill):
    """
    WAGASCI spill number after spill matching with BSD spill
    """

    def __init__(self):
        super(WagasciSpill, self).__init__()
        super(BsdSpill, self).__init__()


class SpillFactory:
    """
    Factory design pattern to produce new spill objects
    """

    def __init__(self):
        pass

    @staticmethod
    def get_spill(type_of_spill):
        # type: (str) -> Union[WagasciSpill, BsdSpill, WagasciBsdSpill]
        """
        Produce a new spill object according to the type of spill
        :param type_of_spill: type of spill object to produce
        :return: spill object
        """
        if type_of_spill.lower() == "wagasci":
            return WagasciSpill()
        elif type_of_spill.lower() == "bsd":
            return BsdSpill()
        elif type_of_spill.lower().replace('_', '').replace(' ', '').replace('-', '') == "wagascibsd":
            return WagasciBsdSpill()
        else:
            raise ValueError("spill type not recognized : {}".format(type_of_spill))


def open_raw_tree(input_file, branch_list, mode="READ"):
    # type: (str, List[str], str) -> Tuple[ROOT.TTree, ROOT.TFile]
    """
    Open a WAGASCI decoded file and set up the TTree for reading the WAGASCI spills.
    Only the needed TBranches are activated for improved performance.
    :param branch_list:
    :param input_file: root file containing WAGASCI decoded data
    :param mode: TFile mode : "READ", "CREATE", "RECREATE", etc...
    :return: TTree and TTfile objects
    """
    tfile = ROOT.TFile.Open(input_file, mode)
    if not tfile:
        raise IOError("ROOT file %s does not exist or is not valid" % input_file)
    print('Opening file "{}"'.format(input_file))

    raw_tree = getattr(tfile, wagascianpy.utils.utils.extract_raw_tree_name(tfile))

    if raw_tree.GetEntries() == 0:
        raise IOError("Raw data TTree is empty in file %s" % input_file)

    if branch_list is not None:
        raw_tree.SetBranchStatus("*", 0)
        for branch in branch_list:
            raw_tree.SetBranchStatus(branch, 1)

    return raw_tree, tfile



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

import os
import subprocess
from typing import Dict, Union, Optional

from six import string_types

import wagascianpy.analysis.analyzer as ana
import wagascianpy.program.program
import wagascianpy.utils.utils


class ProgramBuilder(object):
    """
    The ProgramBuilder interface specifies methods for creating the different parts of
    a program
    """

    def __init__(self, wagasci_lib=None):
        # type: (Optional[str]) -> None
        """
        :param wagasci_lib: path to the directory containing the WAGASCI library
        """
        self._wagasci_lib = wagasci_lib
        self._program = wagascianpy.program.program.Program(self._wagasci_lib)

    @property
    def program(self):
        my_program = self._program
        self.reset()
        return my_program

    def reset(self):
        # type: (...) -> None
        """
        Reset the program object
        :return: None
        """
        self._program = wagascianpy.program.program.Program(self._wagasci_lib)

    def add_decoder(self, **kwargs):
        # type: (...) -> None
        """
        Add the wgDecoder analyzer to the list of analyzers of the program
        :param kwargs: arguments to be passed to wgDecoder
        :return: None
        """
        self._program.add_step(ana.Decoder.name, **kwargs)

    def add_make_hist(self, **kwargs):
        # type: (...) -> None
        """
        Add the wgMakeHist analyzer to the list of analyzers of the program
        :param kwargs: arguments to be passed to wgMakeHist
        :return: None
        """
        self._program.add_step(ana.MakeHist.name, **kwargs)

    def add_tdc_calibration(self, **kwargs):
        # type: (...) -> None
        """
        Add the wgTDC analyzer to the list of analyzers of the program
        :param kwargs: arguments to be passed to wgTDC
        :return: None
        """
        self._program.add_step(ana.TdcCalibration.name, **kwargs)

    def add_bcid_distribution(self, **kwargs):
        # type: (...) -> None
        """
        Create the BCID distribution plots
        :param kwargs: arguments to pass to the BCID distribution analyzer
        :return: None
        """
        self._program.add_step(ana.BcidDistribution.name, **kwargs)

    def add_adc_distribution(self, **kwargs):
        # type: (...) -> None
        """
        Create the ADC distribution plots
        :param kwargs: arguments to pass to the ADC distribution analyzer
        :return: None
        """
        self._program.add_step(ana.AdcDistribution.name, **kwargs)

    def add_spill_number_fixer(self, **kwargs):
        # type: (...) -> None
        """
        Add the wgSpillNumberFixer analyzer to the list of analyzers of the program
        :param kwargs: arguments to be passed to wgSpillNumberFixer
        :return: None
        """
        self._program.add_step(ana.SpillNumberFixer.name, **kwargs)

    def add_beam_summary_data(self,
                              bsd_database_location,
                              bsd_repository_location,
                              download_bsd_database_location,
                              download_bsd_repository_location,
                              t2krun,
                              **kwargs):
        # type: (str, str, str, str, int, ...) -> None
        if ':' in bsd_database_location:
            if len(bsd_database_location.split(':', 1)) != 2:
                raise ValueError("Invalid database location : %s" % bsd_database_location)
            hostname = bsd_database_location.split(':', 1)[0]
            remote_db_path = bsd_database_location.split(':', 1)[-1]
            local_db_path = download_bsd_database_location
            print("Copying remote database %s into location %s" % (remote_db_path, local_db_path))
            wagascianpy.utils.utils.scp_get(hostname, remote_db_path, local_db_path)
        else:
            local_db_path = bsd_database_location
        if not os.path.exists(local_db_path):
            raise EnvironmentError("Local BSD database not found at %s" % local_db_path)

        if ':' in bsd_repository_location:
            if len(bsd_repository_location.split(':', 1)) != 2:
                raise ValueError("Invalid repository location : %s" % bsd_repository_location)
            remote_repo_path = bsd_repository_location
            local_repo_path = download_bsd_repository_location
            print("Copying remote repository %s into location %s" % (remote_repo_path, local_repo_path))
            rsync = wagascianpy.utils.utils.which("rsync")
            if rsync is None:
                raise RuntimeError("rsync program not found")
            try:
                remote = "{}/t2krun{}/*".format(remote_repo_path, t2krun)
                local = "{}/t2krun{}/".format(local_repo_path, t2krun)
                wagascianpy.utils.utils.mkdir_p(local)
                subprocess.check_output([rsync, "-a", "-essh", remote, local])
            except subprocess.CalledProcessError as exception:
                raise RuntimeError("Error while copying the remote repository %s into the "
                                   "local directory %s : %s" % (remote_repo_path, local_repo_path, str(exception)))
        else:
            local_repo_path = bsd_repository_location
        if not os.path.exists(local_repo_path):
            raise EnvironmentError("Local BSD repository not found at %s" % local_repo_path)

        self._program.add_step(ana.BeamSummaryData.name,
                               bsd_database=local_db_path,
                               bsd_repository=local_repo_path,
                               **kwargs)

    def add_apply_detector_flags(self, **kwargs):
        # type: (...) -> None
        """
        Add the apply_detector_flags analyzer to the list of analyzers of the program
        :param kwargs: arguments to be passed to apply_detector_flags
        :return: None
        """
        self._program.add_step(ana.ApplyDetectorFlags.name, **kwargs)

    def add_sanity_check(self, **kwargs):
        # type: (...) -> None
        """
        Add the sanity_check analyzer to the list of analyzers of the program
        :param kwargs: arguments to be passed to sanity_check
        :return: None
        """
        self._program.add_step(ana.SanityCheck.name, **kwargs)

    def add_temperature(self, **kwargs):
        # type: (...) -> None
        """
        Add the wgTemperature analyzer to the list of analyzers of the program
        :param kwargs: arguments to be passed to wgTemperature
        :return: None
        """
        self._program.add_step(ana.Temperature.name, **kwargs)

    def add_adc_calibration(self, **kwargs):
        # type: (...) -> None
        """
        Add the wgAdcCalib analyzer to the list of analyzers of the program
        :param kwargs: arguments to be passed to wgAdcCalib
        :return: None
        """
        self._program.add_step(ana.AdcCalibration.name, **kwargs)

    @property
    def run_location(self):
        # type: (...) -> Dict[str, str]
        """
        Get the input runs dictionary
        :return: input runs dictionary (key: run name, value: run path)
        """
        return self._program.get_run_location()

    @run_location.setter
    def run_location(self, run_dict):
        # type: (Dict[str, str]) -> None
        """
        Set the input runs dictionary
        :param run_dict: input runs dictionary (key: run name, value: run path)
        :return: None
        """
        self._program.set_run_location(run_dict)

    @property
    def save_location(self):
        # type: (...) -> Dict[str, str]
        """
        Get a custom location where to store each run decoded data.
        :return: Dictionary where the key is the run name and the value is the path of the folder where the decoded
                 data is to be stored

        """
        return self._program.get_save_location()

    @save_location.setter
    def save_location(self, save_location):
        # type: (Union[str, Dict[str, str]]) -> None
        """
        Set a custom location where to store each run decoded data.
        :param save_location: Dictionary where the key is the run name and the value is the path of the folder where the
        decoded data is to be stored, or top directory where all the output directories will be created. In case only
        the top directory is specified, the same directory structure of the input runs is recreated for the output.
        :rtype: None
        """
        if not save_location:
            raise ValueError("Save location cannot be empty")
        if not os.path.exists(save_location):
            wagascianpy.utils.utils.mkdir_p(save_location)
        save_dict = {}
        if isinstance(save_location, dict):
            save_dict = save_location
        elif isinstance(save_location, string_types):
            run_dict = self._program.get_run_location()
            for run_name in run_dict:
                save_dict[run_name] = os.path.join(save_location, run_name)
        self._program.set_save_location(save_dict)

    @property
    def multiple_runs_analyzer_save_location(self):
        # type: (...) -> str
        """
        In case of analyzers which need to analyze multiple runs at once and store the results in another directory,
        the user needs to specify it. One example is the wgAdcCalib program that needs to store the gain and dark
        noise history TTrees in some external directory.
        :return: output directory for analyzers that process multiple runs
        """
        return self._program.multiple_runs_analyzer_save_location

    @multiple_runs_analyzer_save_location.setter
    def multiple_runs_analyzer_save_location(self, save_location):
        # type: (str) -> None
        """
        In case of analyzers which need to analyze multiple runs at once and store the results in another directory,
        the user needs to specify it. One example is the wgAdcCalib program that needs to store the gain and dark
        noise history TTrees in some external directory.
        :param save_location: output directory for analyzers that process multiple runs
        :return: None
        """
        self._program.multiple_runs_analyzer_save_location = save_location

    def output_dir_same_as_input(self):
        # type: (...) -> None
        """
        Set the output directory as the same as the input directory
        :return: None
        """
        self._program.output_dir_same_as_input()

    def do_not_stop_on_exception(self):
        # type: (...) -> None
        """
        Do not stop the program execution if an analyzer fails
        :return: None
        """
        self._program._stop_on_exception = False

    def stop_on_exception(self):
        # type: (...) -> None
        """
        Stop the program execution if an analyzer fails
        :return: None
        """
        self._program._stop_on_exception = True

    def enforce_dependencies(self):
        # type: (...) -> None
        """
        Make sure that each analyzer is called in the right order
        :return: None
        """
        self._program._enforce_dependencies = True

    def do_not_enforce_dependencies(self):
        # type: (...) -> None
        """
        Allow an analyzer to be called even if the previous one is not present
        :return: None
        """
        self._program._enforce_dependencies = False



import json
import os

from typing import Mapping, Dict, Optional, Union

import wagascianpy.analysis.analysis
from wagascianpy.utils.environment import WagasciEnvironment

_DETECTORS = {'wallmrd_top_north': {"1": {"1": {"1-3": 32}}},
              'wallmrd_bottom_north': {"1": {"2": {"1-3": 32}}},
              'wallmrd_top_south': {"1": {"3": {"1-3": 32}}},
              'wallmrd_bottom_south': {"1": {"4": {"1-3": 32}}},
              'wagasci_top_upstream': {"2": {"1": {"1-20": 32}}},
              'wagasci_side_upstream': {"2": {"2": {"1-20": 32}}},
              'wagasci_top_downstream': {"2": {"3": {"1-20": 32}}},
              'wagasci_side_downstream': {"2": {"4": {"1-20": 32}}},
              'wallmrd_north': {"1": {"1": {"1-3": 32}, "2": {"1-3": 32}}},
              'wallmrd_south': {"1": {"3": {"1-3": 32}, "4": {"1-3": 32}}},
              'wagasci_upstream': {"2": {"1": {"1-20": 32}, "2": {"1-20": 32}}},
              'wagasci_downstream': {"2": {"3": {"1-20": 32}, "4": {"1-20": 32}}}}
_FULL_SETUP = "wallmrd_north|wallmrd_south|wagasci_upstream|wagasci_downstream"
_MAX_NCHIPS = 20


###############################################################################
#                                 dict_merge                                  #
###############################################################################

def dict_merge(dct, merge_dct):
    # type: (Dict, Dict) -> None
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    for key in merge_dct:
        if (key in dct and isinstance(dct[key], dict)
                and isinstance(merge_dct[key], Mapping)):
            dict_merge(dct[key], merge_dct[key])
        else:
            dct[key] = merge_dct[key]


###############################################################################
#                            parse_topology_string                            #
###############################################################################

def parse_topology_string(topology_string):
    # type: (str) -> str
    """parse topology string containing detector names
        :param topology_string: topology string
        :return parsed topology string
    """
    if topology_string == "full_setup":
        topology_string = _FULL_SETUP
    have_detector_labels = False
    topology_new = {}
    for detector, detector_topology in _DETECTORS.items():
        if detector in topology_string:
            have_detector_labels = True
            dict_merge(topology_new, detector_topology)
    if have_detector_labels:
        topology_old = topology_new.copy()
        topology_new.clear()
    else:
        topology_old = json.loads(topology_string.replace('\\', ''))
        topology_new.clear()
    for gdcc, dif_map in topology_old.items():
        topology_new[gdcc] = {}
        for dif, asu_map_old in dif_map.items():
            asu_key = next(iter(asu_map_old))
            asu_map_new = {}
            if "-" in asu_key:
                asu_list = asu_key.split("-")
                if len(asu_list) != 2:
                    raise ValueError("chip range is invalid : %s" % asu_key)
                n_channels = asu_map_old[asu_key]
                try:
                    first_asu = int(asu_list[0])
                    last_asu = int(asu_list[1])
                except ValueError:
                    raise ValueError("chip range is invalid : %s" % asu_key)
                if first_asu == last_asu or first_asu != 1 or \
                        first_asu is None or last_asu is None or \
                        last_asu - first_asu >= _MAX_NCHIPS:
                    raise ValueError("chip range is invalid : %s" % asu_key)
                for asu in range(first_asu, last_asu + 1, 1):
                    asu_map_new[str(asu)] = n_channels
            else:
                asu_map_new = asu_map_old
            topology_new[gdcc][dif] = asu_map_new
    return json.dumps(topology_new)


###############################################################################
#                read GDCC to DIF topology mapping file                       #
###############################################################################

def read_dif_mapping(mapping_file):
    # type: (Optional[str]) -> Dict[str, Dict[str, int]]
    """ Read the *mapping_file* containing the DIF/GDCC mapping

    # Assuming the following DIF - GDCC mapping as default:
    #
    # GDCC 1 port   -   DIF ID
    # 1                 0
    # 2                 1
    # 3                 2
    # 4                 3
    # GDCC 2 port   -   DIF ID
    # 1                 4
    # 2                 5
    # 3                 6
    # 4                 7
    """
    try:
        with open(mapping_file) as json_file:
            dif_mapping = json.load(json_file)
    except (TypeError, IOError, ValueError) as exception:
        dif_mapping = None
        print('"%s" mapping not found or invalid : %s' % (mapping_file, str(exception)))
    if not dif_mapping:
        print('Using default GDCC-DIF mapping.')
        dif_mapping = json.loads(
            '{"1":{"1":0,"2":1,"3":2,"4":3,"5":8,"6":9,"7":10}'
            ',"2":{"1":4,"2":5,"3":6,"4":7,"5":11,"6":12,"7":13}}')
    return dif_mapping


###############################################################################
#                  read GDCC MAC addresses mapping file                       #
###############################################################################

def read_mac_mapping(mapping_file):
    # type: (str) -> Dict[str, str]
    """ Read the *mapping_file* containing the GDCC to MAC address mapping

    # Assuming the following GDCC MAC mapping as default:
    #
    # GDCC 1
    # 00:0A:35:01:FE:06
    # GDCC 2
    # 00:0A:35:01:FE:01
    """
    try:
        with open(mapping_file) as json_file:
            mac_mapping_inv = json.load(json_file)
            mac_mapping = {value: key for key, value in mac_mapping_inv.items()}
    except (IOError, ValueError) as exception:
        print('"%s" mapping file not found or invalid. '
              'Using default mapping : %s' % (mapping_file, str(exception)))
        mac_mapping = {"1": "00:0A:35:01:FE:06", "2": "00:0A:35:01:FE:01"}
    return mac_mapping


###############################################################################
#                              gdcc2dif_topology                              #
###############################################################################

def gdcc2dif_topology(gdcc_topology, dif_mapping_file=None):
    # type: (Union[str, Dict], Optional[str]) -> Dict
    """
    Convert GDCC topology into DIF topology
    :param gdcc_topology: GDCC topology string
    :param dif_mapping_file: path to the DIF-GDCC port mapping file
    :return: DIF topology
    """
    if isinstance(gdcc_topology, str):
        gdcc_topology = gdcc_topology.replace('\\', '')
        gdcc_topology = json.loads(gdcc_topology)
    dif_mapping = read_dif_mapping(dif_mapping_file)
    dif_topology = {}
    for gdcc, port_dic in gdcc_topology.items():
        for port, chip_dic in port_dic.items():
            dif_id = dif_mapping[gdcc][port]
            dif_topology[str(dif_id)] = {}
            for chip_id, n_channels in chip_dic.items():
                dif_topology[str(dif_id)][str(int(chip_id) - 1)] = n_channels
    return dif_topology


###############################################################################
#                          get_topology_from_xml                              #
###############################################################################

def get_topology_from_xml(acq_config_xml, wagasci_libdir=None):
    # type: (str, Optional[str]) -> Dict
    """
    Generate topology dictionary from the Pyrame configuration XML file
    :param acq_config_xml: path to the Pyrame configuration XML file
    :param wagasci_libdir: path to the WAGASCI libraries directory
    :return: topology dictionary
    """
    env = WagasciEnvironment()
    if wagasci_libdir is None:
        wagasci_libdir = env['WAGASCI_LIBDIR']
    if not os.path.exists(wagasci_libdir):
        raise OSError("WAGASCI library directory not found!")
    wagasci_analyzer = wagascianpy.analysis.analysis.WagasciAnalysis(wagasci_libdir)
    topology_string, pointer = wagasci_analyzer.get_dif_topology(acq_config_xml)
    dif_topology = json.loads(topology_string)
    wagasci_analyzer.free_topology(pointer)
    del wagasci_analyzer
    return dif_topology



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio
#

""" Virtual database record class """

import abc
from datetime import datetime

import pytz
from enum import IntEnum, Enum

# compatible with Python 2 *and* 3:
from typing import Union

ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})

# compatible with Python 2 *and* 3
try:
    # noinspection PyUnresolvedReferences
    IntTypes = (int, long)  # Python2
except NameError:
    IntTypes = int  # Python3

_DEFAULT_TIMEZONE = "Asia/Tokyo"


###########################################################################
#                                DBRecord                                 #
###########################################################################

class DBRecord(ABC):
    """Generic database record record"""

    _timezone = _DEFAULT_TIMEZONE

    def __init__(self, record=None):

        if record is not None:
            for init_member, init_value in record.items():
                if init_member in self.list_fields():
                    setattr(self, init_member, init_value)

    def __getitem__(self, item):
        # type: (str) -> Optional
        return getattr(self, item)

    def is_ready(self):
        """Return True if the make_record method is ready to be called on the object"""
        for member in dir(self):
            value = getattr(self, member)
            if not member.startswith('_') and not callable(value):
                if value is None:
                    print(member)
                    return False
        return True

    def set_timezone(self, timezone):
        """ Change the timezone (default timezone is Asia/Tokyo)"""
        self._timezone = timezone

    def make_record(self):
        """Return a dictionary containing the full record"""
        record = {}
        for member in dir(self):
            value = getattr(self, member)
            if not member.startswith('_') and not callable(value):
                if value is None:
                    raise ValueError("Field %s is not set" % member)
                elif isinstance(value, float):
                    record[member] = float(value)
                elif isinstance(value, (int, Enum, IntEnum)):
                    record[member] = int(value)
                else:
                    record[member] = value
        return record

    def list_fields(self):
        """List all the field names of the record"""
        return list(
            filter(lambda member: not member.startswith('__') and not callable(getattr(self, member)), dir(self)))

    @classmethod
    def datetime2timestamp(cls, datetime_arg):
        # type: (Union[datetime, str, IntTypes, float]) -> float
        """Convert from datetime string to timestamp
        """
        if isinstance(datetime_arg, datetime):
            try:
                try:
                    timestamp = pytz.timezone(DBRecord._timezone).localize(datetime_arg).timestamp()
                except ValueError:
                    timestamp = datetime_arg.timestamp()
            except AttributeError:
                unix_origin = datetime(1970, 1, 1, tzinfo=pytz.utc)
                try:
                    timestamp = (pytz.timezone(DBRecord._timezone).localize(datetime_arg) - unix_origin).total_seconds()
                except ValueError:
                    timestamp = (datetime_arg - unix_origin).total_seconds()
            return float(timestamp)
        elif isinstance(datetime_arg, str):
            return cls.str2timestamp(datetime_arg)
        elif isinstance(datetime_arg, IntTypes) or isinstance(datetime_arg, float):
            return float(datetime_arg)
        else:
            raise ValueError("Datetime argument not recognized %s" % datetime_arg)

    @classmethod
    def timestamp2datetime(cls, posix_timestamp):
        # type: (float) -> datetime
        """Convert from timestamp to datetime
        """
        utc_dt = datetime.utcfromtimestamp(posix_timestamp).replace(tzinfo=pytz.utc)
        return utc_dt.astimezone(pytz.timezone(DBRecord._timezone))

    @classmethod
    def add_timezone(cls, datetime_arg):
        """Add timezone info to datetime
        """
        if datetime_arg.tzinfo is not None and datetime_arg.tzinfo.utcoffset(datetime_arg) is not None:
            return datetime_arg
        else:
            try:
                return datetime_arg.astimezone(pytz.timezone(DBRecord._timezone))
            except ValueError:
                tz = pytz.timezone(DBRecord._timezone)
                return tz.localize(datetime_arg)

    @classmethod
    def timestamp2str(cls, posix_timestamp):
        """Convert from timestamp to human readable date string
        """
        return cls.timestamp2datetime(posix_timestamp).strftime('%Y/%m/%d %H:%M:%S %Z')

    @classmethod
    def str2timestamp(cls, string):
        """Convert from human readable date string to timestamp
        """
        return cls.datetime2timestamp(cls.str2datetime(string))

    @classmethod
    def str2datetime(cls, string):
        # type: (str) -> datetime
        string = string.replace('-', '/')
        if ':' not in string:
            string += ' 00:00:00'
        try:
            time = pytz.timezone(DBRecord._timezone).localize(datetime.strptime(string, "%Y/%m/%d %H:%M:%S"))
        except ValueError:
            time = pytz.timezone(DBRecord._timezone).localize(datetime.strptime(string, "%Y/%m/%d %H:%M:%S %Z"))
        return time

    @classmethod
    def get_timezone_str(cls):
        return cls._timezone

    def pretty_print(self):
        """Print the run record to standard output"""
        for member in dir(self):
            value = getattr(self, member)
            if not member.startswith('_') and not callable(value):
                print("%s : %s" % (member, str(value)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio
#

import argparse
import textwrap
from typing import List


def parse_args(args):
    # type: (List[str]) -> argparse.Namespace
    parser = argparse.ArgumentParser(usage='use "python %(prog)s --help" for more information',
                                     argument_default=None, description=textwrap.dedent('''\
                                     WAGASCI run database and BSD database viewer and manager. If you wish to use the
                                     script arguments from shell, please run in batch mode (option -x). Be carefull 
                                     that all the arguments accept no more than one value and any additional value is 
                                     discarded.'''))

    parser.add_argument('--batch-mode', dest='batch_mode', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        Do not run the graphical interface. (default: %(default)s)'''))

    parser.add_argument('--wagasci-repository', metavar='<WAGASCI repository location>',
                        dest='wagasci_repository', type=str, nargs=1, required=False, help=textwrap.dedent('''\
                        Path to a WAGASCI run repository location. Can be a simple repository or a borg repository.
                        The path can be also a remote path of the form <SSH alias>:/path/to/repository where the SSH
                        alias is the one defined in the .ssh/config file. (default: %(default)s)'''))

    parser.add_argument('--simple-wagasci-repository', dest='simple_wagasci_repository', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        If set the WAGASCI run repository is assumed to be a simple repository (no encription) 
                        (default: %(default)s)'''))

    parser.add_argument('--borg-wagasci-repository', dest='borg_wagasci_repository', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        If set the WAGASCI run repository is assumed to be a borg repository (default: %(default)s)'''))

    parser.add_argument('--update-wagasci-database', dest='update_wagasci_database', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        If set the WAGASCI database is updated (new runs are added but existing runs are not 
                        modified) (default: %(default)s)'''))

    parser.add_argument('--rebuild-wagasci-database', dest='rebuild_wagasci_database', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        If set the WAGASCI database is rebuilt (all runs are overwritten) (default: %(default)s)'''))

    parser.add_argument('--wagasci-database', metavar='<WAGASCI database location>', dest='wagasci_database',
                        type=str, nargs=1, required=False, help=textwrap.dedent('''\
                        Path to the Beam Summary Data database file. This file is created by the 
                        Wagasci Database Viewer (wagasciviewer.py) program and is usually called bsddb.db.
                        The path can be also a remote path of the form <SSH alias>:/path/to/database where the SSH
                        alias is the one defined in the .ssh/config file. (default: %(default)s)
                        '''))

    parser.add_argument('--wagasci-libdir', metavar='<WAGASCI library location>', dest='wagasci_libdir',
                        type=str, nargs=1, required=False, help=textwrap.dedent('''\
                        Path to the WAGASCI library directory. Only local paths are supported. 
                        (default: %(default)s)'''))

    parser.add_argument('--bsd-repository', metavar='<BSD repository location>', dest='bsd_repository', type=str,
                        nargs=1, required=False, help=textwrap.dedent('''
                        Path to the Beam Summary Data local repository. This repository is just a folder 
                        containing all the BSD files. The repository must contain multiple subfolders each one 
                        named t2kXrun where X is the T2K run number. The path can be also a remote path of the form 
                        <SSH alias>:/path/to/repository where the SSH alias is the one defined in the .ssh/config file.
                         (default: %(default)s)'''))

    parser.add_argument('--bsd-database', metavar='<BSD database location>', dest='bsd_database', type=str,
                        nargs=1, required=False, help=textwrap.dedent('''\
                        Path to the Beam Summary Data database file. This file is created by the 
                        Wagasci Database Viewer (wagasciviewer.py) program and is usually called bsddb.db.
                        The path can be also a remote path of the form <SSH alias>:/path/to/database where the SSH
                        alias is the one defined in the .ssh/config file. (default: %(default)s)
                        '''))

    parser.add_argument('--bsd-download-location', metavar='<BSD database download location>',
                        dest='bsd_download_location', type=str, nargs=1, required=False,
                        help=textwrap.dedent('''\
                        Folder where to download BSD files when updating or rebuilding the BSD database
                         (default: %(default)s)
                        '''))

    parser.add_argument('--update-bsd-database', dest='update_bsd_database', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        If set the BSD database is updated (new runs are added but existing runs are not 
                        modified) (default: %(default)s)'''))

    parser.add_argument('--rebuild-bsd-database', dest='rebuild_bsd_database', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        If set the BSD database is rebuilt (all runs are overwritten) (default: %(default)s)'''))

    parser.add_argument('--wagasci-download-location', metavar='<WAGASCI runs download location>',
                        dest='wagasci_download_location', type=str, nargs=1, required=False, help=textwrap.dedent('''
                        Local directory where the WAGASCI runs are downloaded. (default: %(default)s)'''))

    parser.add_argument('--wagasci-decoded-location', metavar='<WAGASCI decoded save location>',
                        dest='wagasci_decoded_location', type=str, nargs=1, required=False, help=textwrap.dedent('''
                        Local directory where the decoded data is stored. (default: %(default)s)'''))

    parser.add_argument('--temperature-sqlite-database', metavar='<temperature SQLite database>',
                        dest='temperature_sqlite_database', type=str, nargs=1, required=False, help=textwrap.dedent('''
                        Local SQLite3 database where all the temperature and humidity readings are stored.
                         (default: %(default)s)'''))

    parser.add_argument('--history-location', metavar='<History location',
                        dest='history_location', type=str, nargs=1, required=False, help=textwrap.dedent('''
                        Local folder where to store the history of gain, noise, etc.. (default: %(default)s)'''))

    parser.add_argument('--t2krun', metavar='<T2K run>', dest='t2krun', type=int,
                        nargs=1, help='T2K run number (default: %(default)s)', default=10, required=False)

    parser.add_argument('--start-time', metavar='<start time>', dest='start_time', type=str,
                        nargs=1, required=False, help=textwrap.dedent('''
                        Start date and time in the form YYYY/MM/DD HH:MM:SS
                        '''))

    parser.add_argument('--stop-time', metavar='<stop time>', dest='stop_time', type=str,
                        nargs=1, required=False, help=textwrap.dedent('''
                        Stop date and time in the form YYYY/MM/DD HH:MM:SS
                        '''))

    parser.add_argument('--start-run', metavar='<start run>', dest='start_run', type=int,
                        nargs=1, default=0, required=False, help=textwrap.dedent('''Start run number'''))

    parser.add_argument('--stop-run', metavar='<stop run>', dest='stop_run', type=int,
                        nargs=1, default=0, required=False, help=textwrap.dedent('''Stop run number'''))

    parser.add_argument('--only-good-runs', dest='only_good_runs', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        If set select only good runs (default: %(default)s)'''))

    parser.add_argument('--include-overlapping', dest='include_overlapping', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        If set include the runs that overlap within the interval (default: %(default)s)'''))

    parser.add_argument('--download', dest='download', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        Download the selected runs. If no run is selected download all runs.'''))

    parser.add_argument('--decoder', dest='decoder', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        Decode the selected runs. If no run is selected decoded all runs.'''))

    parser.add_argument('--make-hist', dest='make_hist', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        Prepare histograms for the selected runs. If no run is selected apply it to all runs.'''))

    parser.add_argument('--tdc-calibration', dest='tdc_calibration', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        Apply TDC calibration to the selected runs. If no run is selected apply it to all runs.'''))

    parser.add_argument('--bcid-distribution', dest='bcid_distribution', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        Plot the BCID histogram for the selected runs. If no run is selected apply it to all runs.'''))

    parser.add_argument('--adc-distribution', dest='adc_distribution', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                            Plot the 2D ADC histogram for the selected runs. 
                            If no run is selected apply it to all runs.'''))

    parser.add_argument('--spill-number-fixer', dest='spill_number_fixer', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        Fix the spill number for the selected runs. If no run is selected apply it to all runs.'''))

    parser.add_argument('--beam-summary-data', dest='beam_summary_data', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        Integrate Beam Summary Data into the selected runs. 
                        If no run is selected apply it to all runs.'''))

    parser.add_argument('--detector-flags', dest='apply_detector_flags', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        Apply detector flags to the selected runs. 
                        If no run is selected apply it to all runs.'''))

    parser.add_argument('--sanity-check', dest='sanity_check', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        Perform a sanity check on the decoded data of to the selected runs. 
                        If no run is selected apply it to all runs.'''))

    parser.add_argument('--adc-calibration', dest='adc_calibration', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        Apply the ADC calibration program to the selected runs. If no run is selected apply it to all 
                        runs, however is recommended to select only the runs from a maintenance day to another, 
                        that is usually one week. This is because after the ADC calibration, usually performed 
                        during the maintenance day, the value of the gain might change slightly.'''))

    parser.add_argument('--temperature', dest='temperature', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        Merge the temperature and humidity for the selected runs. If no run is selected apply it 
                        to all runs. This analyzer must be run after the beam-summary-data analyzer.'''))

    parser.add_argument('--all-analyzers', dest='all_analyzers', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        Apply all analyzers to the selected runs. If no run is selected apply them to all runs.'''))

    parser.add_argument('--overwrite-flag', dest='overwrite_flag', required=False,
                        default=False, action='store_true', help=textwrap.dedent('''\
                        When applying analyzers use the overwrite_flag flag if present'''))

    # Flatten all arguments
    parsed_args = parser.parse_args(args=args)
    for name, value in vars(parsed_args).items():
        if isinstance(value, list):
            setattr(parsed_args, name, value[0])

    return parsed_args
